from __future__ import annotations
from typing import TYPE_CHECKING

import itertools
from contextlib import contextmanager

from trueseeing.core.ui import ui
from trueseeing.core.android.model import Op
from trueseeing.core.android.analyze.op import OpAnalyzer

if TYPE_CHECKING:
  from typing import List, Any, Iterable, Mapping, Set, Optional, FrozenSet, Union, Dict, Iterator
  from typing_extensions import Final
  from trueseeing.core.android.db import APKQuery
  from trueseeing.core.android.model import Token

  DataGraph = Union[Op, Mapping[Op, Any]]

class CodeFlow:
  def __init__(self, q: APKQuery) -> None:
    self._q = q

  def callers_of(self, method_addr: int) -> Iterator[Op]:
    yield from self._q.callers_of(method_addr)

  def callstacks_of(self, method_addr: int) -> Mapping[Op, Any]:
    o = dict()
    for caller in self.callers_of(method_addr):
      o[caller] = self.callstacks_of(caller.addr)
    return o

class DataFlow:
  _q: APKQuery
  _default_max_graph_size: Final[int] = 2 * 1048576
  _an = OpAnalyzer()

  _max_graph_size: int = _default_max_graph_size

  class NoSuchValueError(Exception):
    pass

  class UnsolvableValueError(NoSuchValueError):
    def __init__(self, graph: Optional[DataGraph]) -> None:
      self.graph = graph

  class RegisterDecodeError(Exception):
    pass

  class GraphSizeError(Exception):
    pass

  def __init__(self, q: APKQuery) -> None:
    self._q = q

  @classmethod
  def get_max_graph_size(cls) -> int:
    return cls._max_graph_size

  @classmethod
  def set_max_graph_size(cls, l: Optional[int]) -> int:
    o = cls._max_graph_size
    if l is not None and l != cls._default_max_graph_size:
      ui.info('using graph size limit: {} nodes'.format(l))
      cls._max_graph_size = l
    else:
      cls._max_graph_size = cls._default_max_graph_size
    return o

  @classmethod
  @contextmanager
  def apply_max_graph_size(cls, l: Optional[int]) -> Iterator[None]:
    try:
      o = cls.set_max_graph_size(l)
      yield None
    finally:
      cls.set_max_graph_size(o)

  def likely_calling_in(self, ops: List[Op]) -> None:
    pass

  def into(self, o: Op) -> Optional[DataGraph]:
    return self.analyze(o)

  @classmethod
  def _decoded_registers_of(cls, ref: Token, type_: Any = frozenset) -> Any:
    if ref.t == 'multireg':
      regs = ref.v
      if ' .. ' in regs:
        from_, to_ = regs.split(' .. ')
        return type_([f'{from_[0]}{c:d}' for c in range(int(from_[1:]), int(to_[1:]) + 1)])
      elif ',' in regs:
        return type_([r.strip() for r in regs.split(',')])
      else:
        return type_([regs.strip()])
    elif ref.t == 'reg':
      regs = ref.v
      return type_([regs.strip()])
    elif ref.t == 'reflike' and ref.v == '{},': # XXX
      return type_([])
    else:
      raise cls.RegisterDecodeError(f"unknown type of reference: {ref.t}, {ref.v}")

  @classmethod
  def decoded_registers_of_list(cls, ref: Token) -> List[str]:
    o: List[str] = cls._decoded_registers_of(ref, list)
    return o

  @classmethod
  def decoded_registers_of_set(cls, ref: Token) -> FrozenSet[str]:
    o: FrozenSet[str] = cls._decoded_registers_of(ref, frozenset)
    return o

  # TBD: pack as SQL function
  def looking_behind_from(self, op: Op) -> Iterator[Op]:
    focus = None
    for o in self._q.reversed_insns_in_method(op.addr):
      xs = list(self._an.tokenize(o))
      stem = xs[0]
      if focus is None:
        if stem.t != 'label':
          yield o
        else:
          if not stem.v.startswith("try_"):
            focus = stem.v
      else:
        if stem.t != 'id' or not any(p.v == focus for p in xs[1:]):
          continue
        else:
          focus = None

  def solved_constant_data_in_invocation(self, invokation_op: Op, index: int) -> str:
    assert self._an.get_insn(invokation_op).startswith('invoke')
    graph = self.analyze(invokation_op)
    try:
      reg = self.decoded_registers_of_list(self._an.get_param(invokation_op, 0))[index + (0 if ('-static' in self._an.get_insn(invokation_op)) else 1)]
    except IndexError:
      raise self.NoSuchValueError(f'argument not found at index {index}')
    if graph:
      try:
        arg: Op = graph[invokation_op][reg] # type: ignore
        if self._an.get_insn(arg).startswith('const'):
          return self._an.get_param(arg, 1).v
        else:
          raise self.UnsolvableValueError(graph=arg)
      except (KeyError, AttributeError):
        raise self.UnsolvableValueError(graph=graph)
    else:
      raise self.UnsolvableValueError(graph=None)

  @classmethod
  def walk_dict_values(cls, d: DataGraph) -> Iterable[Optional[Op]]:
    try:
      for v in d.values(): # type: ignore[union-attr]
        yield from cls.walk_dict_values(v)
    except AttributeError:
      yield d # type:ignore[misc]

  def solved_possible_constant_data_in_invocation(self, invokation_op: Op, index: int) -> Set[str]:
    assert self._an.get_insn(invokation_op).startswith('invoke')
    graph = self.analyze(invokation_op)
    reg = self.decoded_registers_of_list(self._an.get_param(invokation_op, 0))[index + (0 if ('-static' in self._an.get_insn(invokation_op)) else 1)]
    if graph:
      n: DataGraph = graph[invokation_op][reg] # type: ignore
      return {self._an.get_param(x, 1).v for x in self.walk_dict_values(n) if x is not None and self._an.get_insn(x).startswith('const')}
    else:
      return set()

  def solved_typeset_in_invocation(self, invokation_op: Op, index: int) -> Set[Any]:
    assert self._an.get_insn(invokation_op).startswith('invoke')
    graph = self.analyze(invokation_op)
    reg = self.decoded_registers_of_list(self._an.get_param(invokation_op, 0))[index + (0 if ('-static' in self._an.get_insn(invokation_op)) else 1)]
    if graph:
      arg: DataGraph = graph[invokation_op][reg] # type: ignore
      return {self._assumed_target_type_of_op(x) for x in self.walk_dict_values(arg) if x is not None}
    else:
      return set()

  @classmethod
  def _assumed_target_type_of_op(cls, x: Op) -> str:
    mn = cls._an.get_insn(x)
    if mn.startswith('const/4'):
      return 'Ljava/lang/Integer;'
    elif mn.startswith('const-string'):
      return 'Ljava/lang/String;'
    elif mn.startswith('new-array'):
      return cls._an.get_param(x, 2).v
    else:
      return 'Ljava/lang/Object;'

  @classmethod
  def _approximated_size_of_graph(cls, d: Optional[DataGraph], *, _cache:Optional[Dict[int, int]] = None) -> int:
    if _cache is None:
      _cache = dict()

    if d is None:
      return 0
    elif isinstance(d, Op):
      return 1
    elif isinstance(d, dict):
      o = 0
      assert len(d) == 1
      for k,v in d.items():
        assert isinstance(k, Op)
        assert k.addr is not None
        if k.addr in _cache:
          return _cache[k.addr]
        else:
          for vk, vv in v.items():
            assert isinstance(vk, str)
            o += cls._approximated_size_of_graph(vv, _cache=_cache)
          _cache[k.addr] = o
      return o
    else:
      assert False

  @classmethod
  def _check_graph(cls, d: Optional[DataGraph]) -> int:
    n = cls._approximated_size_of_graph(d)
    if n > cls._max_graph_size:
      raise cls.GraphSizeError()
    return n

  def analyze(self, op: Optional[Op], state:Optional[Dict[int, Any]]=None, stage:int = 0) -> Optional[DataGraph]:
    if op is None or stage > 64: # XXX: Is it sufficient? Might be better if we make it tunable?
      return None
    if state is None:
      state = dict()

    o: Optional[DataGraph] = None

    mn = self._an.get_insn(op)
    addr = op.addr

    assert addr is not None
    if addr in state:
      return state[addr] # type: ignore[no-any-return]
    try:
      if any(mn.startswith(x) for x in ['const','new-','move-exception']):
        o = op
        state[addr] = o
        return o
      elif mn in ['move', 'array-length']:
        o = {op:{k:self.analyze(self.analyze_recent_load_of(op, k), state, stage=stage+1) for k in self.decoded_registers_of_set(self._an.get_param(op, 1))}}
        ui.debug('analyze: 0x{addr:08x} stage: {stage} ({mode}) -> nodes: {nodes}'.format(mode='move', addr=addr, stage=stage, nodes=self._check_graph(o)))
        state[addr] = o
        return o
      elif any(mn.startswith(x) for x in ['aget-']):
        assert self._an.get_param_count(op) == 3
        o = {op:{k:self.analyze(self.analyze_recent_array_load_of(op, k), state, stage=stage+1) for k in (self.decoded_registers_of_set(self._an.get_param(op, 1)) | self.decoded_registers_of_set(self._an.get_param(op, 2)))}}
        ui.debug('analyze: 0x{addr:08x} stage: {stage} ({mode}) -> nodes: {nodes}'.format(mode='aget', addr=addr, stage=stage, nodes=self._check_graph(o)))
        state[addr] = o
        return o
      elif any(mn.startswith(x) for x in ['sget-']):
        assert self._an.get_param_count(op) == 2
        o = {op:{k:self.analyze(self.analyze_recent_static_load_of(op), state, stage=stage+1) for k in self.decoded_registers_of_set(self._an.get_param(op, 0))}}
        ui.debug('analyze: 0x{addr:08x} stage: {stage} ({mode}) -> nodes: {nodes}'.format(mode='sget', addr=addr, stage=stage, nodes=self._check_graph(o)))
        state[addr] = o
        return o
      elif any(mn.startswith(x) for x in ['iget-']):
        assert self._an.get_param_count(op) == 3
        o = {op:{k:self.analyze(self.analyze_recent_instance_load_of(op), state, stage=stage+1) for k in self.decoded_registers_of_set(self._an.get_param(op, 0))}}
        ui.debug('analyze: 0x{addr:08x} stage: {stage} ({mode}) -> nodes: {nodes}'.format(mode='iget', addr=addr, stage=stage, nodes=self._check_graph(o)))
        state[addr] = o
        return o
      elif mn.startswith('move-result'):
        o = self.analyze(self.analyze_recent_invocation(op), state, stage=stage+1)
        ui.debug('analyze: 0x{addr:08x} stage: {stage} ({mode}) -> nodes: {nodes}'.format(mode='move-result', addr=addr, stage=stage, nodes=self._check_graph(o)))
        state[addr] = o
        return o
      else:
        try:
          o = {op:{k:self.analyze(self.analyze_recent_load_of(op, k), state, stage=stage+1) for k in self.decoded_registers_of_set(self._an.get_param(op, 0))}}
          ui.debug('analyze: 0x{addr:08x} stage: {stage} ({mode}) -> nodes: {nodes}'.format(mode='gen.', addr=addr, stage=stage, nodes=self._check_graph(o)))
          state[addr] = o
          return o
        except self.RegisterDecodeError:
          state[addr] = None
          return None
    except self.GraphSizeError:
      ui.warn('analyze: 0x{addr:08x} stage: {stage}: too many nodes'.format(addr=addr, stage=stage))
      state[addr] = None
      return None

  def analyze_recent_static_load_of(self, op: Op) -> Optional[Op]:
    assert self._an.get_insn(op).startswith('sget-')
    target = self._an.get_param(op, 1).v
    for o in itertools.chain(self.looking_behind_from(op), self._q.sputs(target)):
      try:
        if self._an.get_insn(o).startswith('sput-'):
          if self._an.get_param(o, 1).v == target:
            return o
      except ValueError:
        pass
    else:
      ui.debug(f"analyze_recent_static_load_of: failed static trace: {op!r}")
      return None

  def analyze_load(self, op: Op) -> FrozenSet[str]:
    try:
      v = self._an.get_insn(op)
    except ValueError:
      return frozenset()
    else:
      if any(v.startswith(x) for x in ['const','new-','move','array-length','aget-','sget-','iget-']):
        return self.decoded_registers_of_set(self._an.get_param(op, 0))
      elif any(v.startswith(x) for x in ['invoke-direct', 'invoke-virtual', 'invoke-interface']):
        # Imply modification of "this"
        return frozenset(self.decoded_registers_of_list(self._an.get_param(op, 0))[:1])
      return frozenset()

  def analyze_recent_load_of(self, from_: Op, reg: str, stage: int = 0) -> Optional[Op]:
    for o in self.looking_behind_from(from_):
      if self._an.get_mnemonic(o).t == 'id':
        if reg in self.analyze_load(o):
          return o
    if reg.startswith('p'):
      index = int(reg.replace('p', ''))
      for caller in CodeFlow(self._q).callers_of(from_.addr):
        if not self._q.in_same_mod(from_.addr, caller.addr):
          caller_reg = self.decoded_registers_of_list(self._an.get_param(caller, 0))[index]
          if ui.is_debugging:
            ui.debug('analyze_recent_load_of: retracing [{stage}]: {reg} 0x{addr:08x} [{qn}] -> {caller_reg} 0x{caller_addr:08x} [{caller_qn}] {{{caller_l}}}'.format(
              stage=stage, reg=reg, addr=from_.addr, qn=self._q.qualname_of(from_.addr),
              caller_reg=caller_reg, caller_addr=caller.addr, caller_qn=self._q.qualname_of(caller.addr),
              caller_l=self._q.op_get(caller.addr).l,
            ))
          if stage < 5:
            retraced = self.analyze_recent_load_of(caller, caller_reg, stage=stage+1)
            if retraced:
              return retraced
    return None

  # TBD: tracing on static-array fields
  def analyze_recent_array_load_of(self, from_: Op, reg: str) -> Optional[Op]:
    return self.analyze_recent_load_of(from_, reg)

  # TBD: tracing on static-instance fields
  def analyze_recent_instance_load_of(self, op: Op) -> Optional[Op]:
    assert self._an.get_param_count(op) == 3
    assert self._an.get_insn(op).startswith('iget-')
    field = self._an.get_param(op, 2).v
    for o in itertools.chain(self.looking_behind_from(op), self._q.iputs(field)):
      try:
        if self._an.get_insn(o).startswith('iput-') and self._an.get_param(o, 2).v == field:
          return o
      except ValueError:
        pass
    return None

  def analyze_recent_invocation(self, from_: Op) -> Optional[Op]:
    for o in self.looking_behind_from(from_):
      try:
        if self._an.get_insn(o).startswith('invoke'):
          return o
      except ValueError:
        pass
    return None



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
  from typing import TypedDict

  class Call(TypedDict):
    path: str
    sect: str
    offs: int
    priv: bool
    swift: bool
    objc: bool
    cpp: bool
    target: str



from __future__ import annotations
from typing import TYPE_CHECKING

import lxml.etree as ET
import os
import re
from functools import cache

from pubsub import pub

from trueseeing.core.ui import ui
from trueseeing.core.env import get_cache_dir, get_cache_dir_v0, get_cache_dir_v1
from trueseeing.core.context import Context, Fingerprint

if TYPE_CHECKING:
  from typing import List, Any, Iterable, Tuple, Optional, ClassVar, Set, AsyncIterator, Iterator, Mapping
  from trueseeing.core.context import ContextType, ContextInfo
  from trueseeing.core.android.asm import APKDisassembler
  from trueseeing.core.android.store import APKStore
  from trueseeing.core.android.model import XAPKManifest, Call

class PackageNameReader:
  @cache
  def read(self, path: str) -> str:
    if path.endswith('.apk'):
      from pyaxmlparser import APK
      apk = APK(path)
      return apk.packagename # type: ignore[no-any-return]
    elif path.endswith('.xapk'):
      from zipfile import ZipFile
      with ZipFile(path) as zf:
        from json import loads
        manif: XAPKManifest = loads(zf.read('manifest.json'))
        vers = manif['xapk_version']
        if str(vers) != '2':
          raise ValueError(f'invalid xapk manifest: {vers}')
        pkg = manif['package_name']
        return pkg
    else:
      raise ValueError('format unknown')

class APKContext(Context):
  _store: Optional[APKStore] = None
  _type: ClassVar[Set[ContextType]] = {'apk', 'file'}
  _package_reader = PackageNameReader()
  _fp = Fingerprint()

  def invalidate(self) -> None:
    super().invalidate()
    if self._store:
      self._store.invalidate()
      self._store = None
    self.disassembled_classes.cache_clear()
    self.disassembled_resources.cache_clear()
    self.disassembled_assets.cache_clear()
    self._string_resource_files.cache_clear()
    self._xml_resource_files.cache_clear()
    self._fp.get.cache_clear()

  def _get_workdir(self) -> str:
    path = self._get_workdir_v3()
    if os.path.isdir(path):
      return path
    else:
      fp = self._get_fingerprint()
      paths = dict(v0=self._get_workdir_v0(fp), v1=self._get_workdir_v1(fp), v2=self._get_workdir_v2(fp))
      for vers,trypath in paths.items():
        if os.path.isdir(trypath):
          ui.warn(f'cpntext uses old path ({vers}), consider moving it to {path}')
          return trypath
      else:
        return path

  def _get_workdir_v3(self) -> str:
    from hashlib import sha256
    ctx_id = sha256((self._path + ':' + self.get_package_name()).encode('utf-8')).hexdigest()
    return os.path.join(get_cache_dir(), ctx_id)

  def _get_workdir_v2(self, fp: str) -> str:
    return os.path.join(get_cache_dir(), fp)

  def _get_workdir_v1(self, fp: str) -> str:
    return os.path.join(get_cache_dir_v1(self._path), f'.trueseeing2-{fp}')

  def _get_workdir_v0(self, fp: str) -> str:
    return os.path.join(get_cache_dir_v0(), fp)

  def store(self) -> APKStore:
    if self._store is None:
      assert self.wd is not None
      from trueseeing.core.android.store import APKStore
      self._store = APKStore(self.wd)
    return self._store

  def _get_type(self) -> Set[ContextType]:
    return self._type

  def _get_size(self) -> Optional[int]:
    return os.stat(self._path).st_size

  def _get_fingerprint(self) -> str:
    return self._fp.get(self._path)

  async def _recheck_schema(self) -> None:
    from trueseeing.core.android.store import APKStore
    APKStore.require_valid_schema_on(self.wd)

  async def _get_disassembler(self) -> APKDisassembler:
    from trueseeing.core.android.asm import APKDisassembler
    return APKDisassembler(self)

  async def _analyze(self, level: int) -> None:
    from time import time

    at = time()

    await self._analyze_dalvik(level)
    await self._analyze_native(level)

    pub.sendMessage('progress.core.analysis.done', t=time()-at)

  async def _analyze_dalvik(self, level: int) -> None:
    pub.sendMessage('progress.core.context.disasm.begin')
    disasm = await self._get_disassembler()
    await disasm.disassemble(level)
    pub.sendMessage('progress.core.context.disasm.done')

    if level > 3:
      import time
      from io import StringIO

      started = time.time()
      pat = 'smali/%.smali'
      with self.store().query().scoped() as q:
        total = q.file_count(pat)
        pub.sendMessage('progress.core.analysis.smali.begin', total=total)
        with q.db as c:
          def _addr(nr: int, ln: int) -> int:
            return (nr << 24 | ln)

          def _addr_floor(nr: int) -> int:
            return _addr(nr, 0)

          def _addr_ceil(nr: int) -> int:
            return _addr_floor(nr+1) - 1

          nr = 0
          c.execute('create temporary table _tmp_ops (addr integer not null primary key, l varchar not null)')
          c.execute('create temporary table _tmp_ops_map (addr integer primary key, map_id integer)')
          c.execute('create temporary table _tmp_xref (addr integer not null, insn varchar not null, val varchar not null, unique (addr, insn, val))')

          for n, b in q.file_enum(pat):
            t = b.decode('utf-8')
            m = re.search(r'^\.class.* (L.*?;)', t)
            assert m is not None, t
            class_name = m.group(1)
            m = re.search(r'^\.super (L.*?;)', t, re.MULTILINE)
            assert m is not None
            extends = m.group(1)
            implements = [m.group(1) for m in re.finditer(r'^\.implements (L.*?;)$', t, re.MULTILINE)]
            c.execute('insert into class_rel(class, super) values (:class_name, :extends)', dict(class_name=class_name, extends=extends))
            for impl in implements:
              c.execute('insert into class_rel(class, impl) values (:class_name, :impl)', dict(class_name=class_name, impl=impl))

            c.execute('insert into map(low, high, class) values (:low, :high, :class_name)', dict(low=_addr_floor(nr), high=_addr_ceil(nr), class_name=class_name))
            method_seen: Optional[Tuple[int, str]] = None
            for ln, l in enumerate(StringIO(t)):
              addr = _addr(nr, ln)
              l = l.rstrip('\n')
              if not l or ' .line ' in l or l.startswith('.source '):
                continue
              c.execute('insert into _tmp_ops(addr, l) values (:addr, :l)', dict(addr=addr, l=l))
              if not method_seen:
                m = re.search(r'^\.method .*? ([^ ]+)$', l)
                if m:
                  method_seen = ln, m.group(1)
              else:
                m = re.search(r'^  +((?:invoke|[si]put|const)\S+)', l)
                if m:
                  insn = m.group(1)
                  mv = re.search(r', (".*"|\S+)(?: +# .+?)?$', l)
                  assert mv, l
                  c.execute('insert into _tmp_xref (addr, insn, val) values (:addr, :insn, :val)', dict(addr=addr, insn=insn, val=mv.group(1)))
                else:
                  if l.startswith(r'.end method'):
                    start, name = method_seen
                    c.execute('insert into map(low, high, class, method) values (:low, :high, :class_name, :method_name)', dict(low=_addr(nr, start), high=addr, class_name=class_name, method_name=name))
                    method_seen = None

            nr += 1
            pub.sendMessage('progress.core.analysis.smali.analyzing', nr=nr)

          pub.sendMessage('progress.core.analysis.smali.analyzed')

          nr_ops = 0
          for nr, in c.execute('select count(1) from _tmp_ops'):
            nr_ops = nr
            pub.sendMessage('progress.core.analysis.smali.summary', ops=nr)

          nr_classes = 0
          for nr, in c.execute('select count(1) from map where method is null'):
            nr_classes = nr
            pub.sendMessage('progress.core.analysis.smali.summary', ops=nr_ops, classes=nr_classes)

          nr_methods = 0
          for nr, in c.execute('select count(1) from map where method is not null'):
            nr_methods = nr
            pub.sendMessage('progress.core.analysis.smali.summary', ops=nr_ops, classes=nr_classes, methods=nr_methods)

          pub.sendMessage('progress.core.analysis.smali.finalizing')

          for addr, insn, val in c.execute('select addr, insn, val from _tmp_xref'):
            if insn.startswith('const'):
              c.execute('insert into xref_const (addr, insn, sym) values (:addr, :insn, :sym)', dict(addr=addr, insn=insn, sym=val if not val.startswith('"') else val[1:-1]))
            else:
              nc = val.split('->')
              assert len(nc) == 2
              if insn.startswith('invoke'):
                assert '(' in nc[1]
                c.execute('insert into xref_invoke (addr, insn, sym, target) values (:addr, :insn, :sym, (select low from map where class=:cn and method=:mn))', dict(addr=addr, insn=insn, sym=val, cn=nc[0], mn=nc[1]))
              else:
                assert '(' not in nc[1]
                if insn.startswith('sput'):
                  c.execute('insert into xref_sput (addr, insn, sym) values (:addr, :insn, :sym)', dict(addr=addr, insn=insn, sym=val))
                elif insn.startswith('iput'):
                  c.execute('insert into xref_iput (addr, insn, sym) values (:addr, :insn, :sym)', dict(addr=addr, insn=insn, sym=val))

          c.execute('insert into _tmp_ops_map select A.addr,B.id from _tmp_ops as A join map as B on (A.addr between B.low and B.high) where B.method is null')
          c.execute('insert or replace into _tmp_ops_map select A.addr,B.id from _tmp_ops as A join map as B on (A.addr between B.low and B.high) where B.method is not null')
          c.execute('insert into ops (addr, l, map_id) select addr, l, map_id from _tmp_ops join _tmp_ops_map using (addr)')
          c.execute('analyze')
          pub.sendMessage('progress.core.analysis.smali.done', t=time.time() - started)

  async def _analyze_native(self, level: int) -> None:
    if level > 2:
      tarpath = os.path.join(os.path.dirname(self._path), 'disasm.tar.gz')
      if not os.path.exists(tarpath):
        ui.warn(f'skipping native code analysis; prepare {tarpath}')
        return

      from time import time
      at = time()

      with self.store().query().scoped() as q:
        pub.sendMessage('progress.core.analysis.nat.begin')
        import tarfile
        with tarfile.open(tarpath) as tf:
          q.file_put_batch(dict(path=i.name, blob=tf.extractfile(i).read(), z=True) for i in tf.getmembers() if (i.isreg() or i.islnk())) # type:ignore[union-attr]

        if level > 3:
          pub.sendMessage('progress.core.analysis.nat.analyzing')
          from trueseeing.core.android.analyze.nat import analyze_api_in

          def _as_call(g: Iterator[Mapping[str, Any]]) -> Iterator[Call]:
            for e in g:
              typ = e['typ']
              lang = e['lang']
              sect, offs = e['origin'].split('+')
              yield dict(
                path=e['fn'],
                sect=sect,
                offs=int(offs.strip(), 16),
                priv=(typ == 'private'),
                cpp=(lang == 'cpp'),
                target=e['call']
              )

          q.call_add_batch(_as_call(analyze_api_in(q.file_enum('lib/%'))))
          pub.sendMessage('progress.core.analysis.nat.summary', calls=q.call_count())

      pub.sendMessage('progress.core.analysis.nat.done', t=time() - at)

  def get_package_name(self) -> str:
    return self._package_reader.read(self.target)

  async def _get_info(self, extended: bool) -> AsyncIterator[ContextInfo]:
    async for m in super()._get_info(extended):
      yield m

    level = self.get_analysis_level()
    if level > 0:
      store = self.store()
      manif = self.parsed_manifest()
      yield dict(
        pkg=self.get_package_name(),
        ver='{} ({})'.format(
          manif.attrib['{http://schemas.android.com/apk/res/android}versionName'],
          manif.attrib['{http://schemas.android.com/apk/res/android}versionCode']
        ),
        perms=len(list(self.permissions_declared())),
        activs=len(list(manif.xpath('.//activity'))),
        servs=len(list(manif.xpath('.//service'))),
        recvs=len(list(manif.xpath('.//receiver'))),
        provs=len(list(manif.xpath('.//provider'))),
      )
      yield {'int-flts':len(list(manif.xpath('.//intent-filter')))}
      if level > 3:
        with store.db as c:
          for nr, in c.execute('select count(1) from class_rel where super regexp :pat', dict(pat='^Landroid.*Fragment(Compat)?;$')):
            yield dict(frags=nr)
      for e in manif.xpath('.//application'):
        boolmap = {True:'true', False:'false', 'true':'true', 'false':'false'}
        yield {
          'debuggable?':boolmap.get(e.attrib.get('{http://schemas.android.com/apk/res/android}debuggable', 'false'), '?'),
          'backupable?':boolmap.get(e.attrib.get('{http://schemas.android.com/apk/res/android}allowBackup', 'false'), '?'),
          'netsecconf?':boolmap.get(e.attrib.get('{http://schemas.android.com/apk/res/android}networkSecurityConfig') is not None, '?'),
        }
      if manif.xpath('.//uses-sdk'):
        for e in manif.xpath('.//uses-sdk'):
          yield {
            'api min':int(e.attrib.get('{http://schemas.android.com/apk/res/android}minSdkVersion', '1')),
            'api tgt':int(e.attrib.get('{http://schemas.android.com/apk/res/android}targetSdkVersion', '1')),
          }
      else:
        ui.warn('cannot determine min/target sdk version')
        yield {
          'api min':'?',
          'api tgt':'?',
        }
      if level == 3:
        with store.query().scoped() as q:
          yield dict(classes='~{}'.format(q.file_count('smali/%')))
      elif level > 3:
        with store.db as c:
          for nr, in c.execute('select count(1) from analysis_issues'):
            yield dict(issues='{}{}'.format(nr, ('' if nr else ' (not scanned yet?)')))
          for nr, in c.execute('select count(1) from ops'):
            yield dict(ops='{}'.format(nr))
          for nr, in c.execute('select count(1) from map where method is null'):
            yield dict(classes='{}'.format(nr))
          for nr, in c.execute('select count(1) from map where method is not null'):
            yield dict(methods='{}'.format(nr))

    if extended:
      from subprocess import CalledProcessError
      from trueseeing.core.android.device import AndroidDevice
      dev = AndroidDevice()
      try:
        build = (await dev.invoke_adb('shell getprop ro.build.fingerprint', catch_stderr=True)).rstrip()
        yield {
          'device?': 'yes ({})'.format(build),
        }
      except CalledProcessError as e:
        cause = e.stderr.decode().splitlines()[-1]
        if cause.startswith('adb: no '):
          yield {
            'device?': 'no',
          }
        else:
          yield {
            'device?': '? ({})'.format(cause)
          }

  def parsed_manifest(self, patched: bool = False) -> Any:
    return self.store().query().file_get_xml('AndroidManifest.xml', patched=patched)

  def manifest_as_xml(self, manifest: Any) -> bytes:
    assert manifest is not None
    return ET.tostring(manifest) # type: ignore[no-any-return]

  def get_target_sdk_version(self) -> int:
    manif = self.parsed_manifest()
    e = manif.xpath('.//uses-sdk')[0]
    return int(e.attrib.get('{http://schemas.android.com/apk/res/android}targetSdkVersion', '1'))

  # FIXME: Handle invalid values
  def get_min_sdk_version(self) -> int:
    manif = self.parsed_manifest()
    e = manif.xpath('.//uses-sdk')[0]
    return int(e.attrib.get('{http://schemas.android.com/apk/res/android}minSdkVersion', '1'))

  @cache
  def disassembled_classes(self) -> List[str]:
    return list(self.store().query().file_find('smali%.smali'))

  @cache
  def disassembled_resources(self) -> List[str]:
    return list(self.store().query().file_find('%/res/%.xml'))

  @cache
  def disassembled_assets(self) -> List[str]:
    return list(self.store().query().file_find('root/%/assets/%'))

  def source_name_of_disassembled_class(self, fn: str) -> str:
    return os.path.join(*fn.split('/')[2:])

  def dalvik_type_of_disassembled_class(self, fn: str) -> str:
    return 'L{};'.format((self.source_name_of_disassembled_class(fn).replace('.smali', '')))

  def source_name_of_disassembled_resource(self, fn: str) -> str:
    return os.path.join(*fn.split('/')[3:])

  def class_name_of_dalvik_class_type(self, dc: str) -> str:
    return re.sub(r'^L|;$', '', dc).replace('/', '.')

  def permissions_declared(self) -> Iterable[Any]:
    yield from self.parsed_manifest().xpath('//uses-permission/@android:name', namespaces=dict(android='http://schemas.android.com/apk/res/android'))

  @cache
  def _string_resource_files(self) -> List[str]:
    return list(self.store().query().file_find('%/res/values/%strings%'))

  def string_resources(self) -> Iterable[Tuple[str, str]]:
    for _, o in self.store().query().file_enum('%/res/values/%strings%'):
      yield from ((c.attrib['name'], c.text) for c in ET.fromstring(o, parser=ET.XMLParser(recover=True)).xpath('//resources/string') if c.text)

  @cache
  def _xml_resource_files(self) -> List[str]:
    return list(self.store().query().file_find('%/res/xml/%.xml'))

  def xml_resources(self) -> Iterable[Tuple[str, Any]]:
    for fn, o in self.store().query().file_enum('%/res/xml/%.xml'):
      yield (fn, ET.fromstring(o, parser=ET.XMLParser(recover=True)))

  def is_qualname_excluded(self, qualname: Optional[str]) -> bool:
    if qualname is not None:
      return any([re.match(f'L{x}', qualname) for x in self.excludes])
    else:
      return False

class XAPKContext(APKContext):
  _disasm: Optional[APKDisassembler] = None
  _type: ClassVar[Set[ContextType]] = {'xapk'}

  async def _get_disassembler(self) -> APKDisassembler:
    assert self._disasm
    return self._disasm

  async def _analyze(self, level: int) -> None:
    from shlex import quote
    from tempfile import TemporaryDirectory
    with TemporaryDirectory(dir=self.wd) as td:
      from trueseeing.core.tools import invoke_streaming
      from trueseeing.core.android.tools import toolchains
      outfile = os.path.join(td, 'merged.apk')
      with toolchains() as tc:
        async for l in invoke_streaming('java -jar {apkeditor} m -i {target} -o {outfile}'.format(
            apkeditor=tc['apkeditor'],
            target=quote(self.target),
            outfile=outfile,
        )):
          ui.info(l.decode())
      try:
        from trueseeing.core.android.asm import APKDisassembler
        self._disasm = APKDisassembler(self, outfile)
        return await super()._analyze(level)
      finally:
        self._disasm = None

  def _get_type(self) -> Set[ContextType]:
    return super()._type | self._type

  async def _get_info(self, extended: bool) -> AsyncIterator[ContextInfo]:
    async for m in super()._get_info(extended):
      yield m

    manif = self._get_xapk_manifest()
    yield {
      'xapk vers': manif['xapk_version'],
      'xapk slices': '{} ({})'.format(len(manif['split_apks']), ', '.join([x['id'] for x in manif['split_apks']])),
    }

  def _get_xapk_manifest(self) -> XAPKManifest:
    from zipfile import ZipFile
    from json import loads
    with ZipFile(self.target) as zf:
      manif: XAPKManifest = loads(zf.read('manifest.json'))
      assert str(manif['xapk_version']) == '2'
      return manif



from __future__ import annotations
from typing import TYPE_CHECKING

import sys
from contextlib import contextmanager
from functools import cache
from pubsub import pub
from trueseeing.core.exc import FatalError

if TYPE_CHECKING:
  from typing import NoReturn, Optional, TextIO, Set, Any, Iterator, Iterable, Tuple
  from typing_extensions import Final
  from progressbar import ProgressBar
  from trueseeing.core.model.issue import Issue
  from trueseeing.core.android.model import Op
  from trueseeing.core.android.db import APKQuery

class UI:
  DEBUG: Final = 0
  INFO: Final = 1
  WARN: Final = 2
  ERROR: Final = 3
  CRITICAL: Final = 4
  FATAL: Final = 5

  level = DEBUG
  is_debugging = False
  _is_inspecting = False

  _seen: Set[str] = set()

  def is_tty(self, stdin: bool = False) -> bool:
    from os import isatty
    if not stdin:
      return isatty(sys.stderr.fileno())
    else:
      return isatty(sys.stdin.fileno())

  def enter_inspect(self) -> None:
    self._is_inspecting = True

  def exit_inspect(self) -> None:
    self._is_inspecting = False

  def set_level(self, level: int) -> None:
    self.level = level
    self.is_debugging = (self.level == self.DEBUG)

  # XXX: check color capability on our own because we are coloring stderr -- termcolor cares stdout only.
  @cache
  def colored(self, x: str, **kw: Any) -> str:
    from termcolor import colored
    return colored(x, force_color=self._can_do_colour_stderr(), **kw)

  def bullet(self, what: str) -> str:
    if not self._is_inspecting:
      return ''
    if what == 'critical':
      return self.colored('[!] ', color='red', attrs=('bold',))
    elif what == 'error':
      return self.colored('[-] ', color='red', attrs=('bold',))
    elif what == 'warn':
      return self.colored('[*] ', color='yellow', attrs=('bold',))
    elif what == 'info':
      return self.colored('[*] ', color='blue', attrs=('bold',))
    elif what == 'debug':
      return self.colored('[.] ', color='grey', attrs=('bold',))
    elif what == 'success':
      return self.colored('[+] ', color='green', attrs=('bold',))
    elif what == 'failure':
      return self.colored('[-] ', color='red', attrs=('bold',))
    assert False, f'invalid type of bullet: {what}'

  def fatal(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> NoReturn:
    if not self._is_inspecting:
      self.stderr(f'fatal: {msg}', nl=nl, ow=ow, onetime=onetime, exc=exc)
    else:
      self.failure(f'fatal: {msg}', nl=nl, ow=ow, onetime=onetime, exc=exc)
    raise FatalError()

  def critical(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    if self.level <= self.CRITICAL:
      self.stderr(self._format_msg(msg, 'critical'), nl=nl, ow=ow, onetime=onetime, exc=exc)

  def error(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    if self.level <= self.ERROR:
      self.stderr(self._format_msg(msg, 'error'), nl=nl, ow=ow, onetime=onetime, exc=exc)

  def warn(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    if self.level <= self.WARN:
      self.stderr(self._format_msg(msg, 'warn'), nl=nl, ow=ow, onetime=onetime, exc=exc)

  def info(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    if self.level <= self.INFO:
      self.stderr(self._format_msg(msg, 'info'), nl=nl, ow=ow, onetime=onetime, exc=exc)

  def debug(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    if self.level <= self.DEBUG:
      self.stderr(self._format_msg(msg, 'debug'), nl=nl, ow=ow, onetime=onetime, exc=exc)

  def success(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    self.stderr(self._format_msg(msg, 'success'), nl=nl, ow=ow, onetime=onetime, exc=exc)

  def failure(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    self.stderr(self._format_msg(msg, 'failure'), nl=nl, ow=ow, onetime=onetime, exc=exc)

  def stdout(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    if onetime:
      if msg in self._seen:
        return
      else:
        self._seen.add(msg)
    if ow:
      sys.stdout.write('\r')
    sys.stdout.write(msg)
    if nl:
      sys.stdout.write('\n')
    sys.stdout.flush()
    if exc is not None:
      self._format_exception(sys.stdout, exc, nl=nl, ow=ow)

  def stderr(self, msg: str, nl: bool = True, ow: bool = False, onetime: bool = False, exc: Optional[BaseException] = None) -> None:
    if onetime:
      if msg in self._seen:
        return
      else:
        self._seen.add(msg)
    if ow:
      sys.stderr.write('\r')
    sys.stderr.write(msg)
    if nl:
      sys.stderr.write('\n')
    sys.stderr.flush()
    if exc is not None:
      self._format_exception(sys.stderr, exc, nl=nl, ow=ow)

  def _format_exception(self, f: TextIO, exc: BaseException, nl: bool = True, ow: bool = False) -> None:
    from traceback import format_exception
    if ow:
      f.write('\r')
    f.write(''.join(format_exception(type(exc), exc, exc.__traceback__)))
    if nl:
      f.write('\n')

  def _format_msg(self, msg: str, flagtyp: str, **kw: Any) -> str:
    return '{flag}{msg}'.format(flag=self.bullet(flagtyp), msg=msg)

  # termcolor 2.4 compatible color capability checker
  @cache
  def _can_do_colour_stderr(self) -> bool:
    from io import UnsupportedOperation
    from os import environ, isatty

    if "ANSI_COLORS_DISABLED" in environ:
      return False
    if "NO_COLOR" in environ:
      return False
    if "FORCE_COLOR" in environ:
      return True

    if environ.get("TERM") == "dumb":
      return False
    if not hasattr(sys.stderr, "fileno"):
      return False

    try:
      return isatty(sys.stderr.fileno())
    except UnsupportedOperation:
      return sys.stderr.isatty()

class CoreProgressReporter:
  _bar_lift: Optional[ProgressBar] = None
  _bar_disasm: Optional[ProgressBar] = None
  _bar_asm: Optional[ProgressBar] = None
  _bar_analysis: Optional[ProgressBar] = None

  @contextmanager
  def scoped(self) -> Iterator[None]:
    submap = {
      'progress.core.context.disasm.begin':self._core_context_disasm_begin,
      'progress.core.context.disasm.done':self._core_context_disasm_done,
      'progress.core.asm.lift.begin':self._core_asm_lift_begin,
      'progress.core.asm.lift.update':self._core_asm_lift_update,
      'progress.core.asm.lift.done':self._core_asm_lift_done,
      'progress.core.asm.disasm.begin':self._core_asm_disasm_begin,
      'progress.core.asm.disasm.update':self._core_asm_disasm_update,
      'progress.core.asm.disasm.done':self._core_asm_disasm_done,
      'progress.core.asm.asm.begin':self._core_asm_asm_begin,
      'progress.core.asm.asm.update':self._core_asm_asm_update,
      'progress.core.asm.asm.done':self._core_asm_asm_done,
      'progress.core.analysis.smali.begin':self._core_analysis_smali_begin,
      'progress.core.analysis.smali.analyzing':self._core_analysis_smali_analyzing,
      'progress.core.analysis.smali.analyzed':self._core_analysis_smali_analyzed,
      'progress.core.analysis.smali.summary':self._core_analysis_smali_summary,
      'progress.core.analysis.smali.finalizing':self._core_analysis_smali_finalizing,
      'progress.core.analysis.smali.done':self._core_analysis_smali_done,
      'progress.core.analysis.nat.begin':self._core_analysis_nat_begin,
      'progress.core.analysis.nat.analyzing':self._core_analysis_nat_analyzing,
      'progress.core.analysis.nat.summary':self._core_analysis_nat_summary,
      'progress.core.analysis.nat.done':self._core_analysis_nat_done,
      'progress.core.analysis.done':self._core_analysis_done,
    }
    try:
      for k, v in submap.items():
        pub.subscribe(v, k)
      yield None
    finally:
      for k, v in submap.items():
        pub.unsubscribe(v, k)
      pass

  def _core_context_disasm_begin(self) -> None:
    ui.info('analyze: disassembling... ', nl=False)

  def _core_asm_lift_begin(self) -> None:
    if ui.is_tty():
      from progressbar import ProgressBar, RotatingMarker
      self._bar_lift = ProgressBar(widgets=[
        ui.bullet('info'),
        'analyze: disassembling... ',
        RotatingMarker()   # type:ignore[no-untyped-call]
      ])
    else:
      self._bar_lift = None

  def _core_asm_lift_update(self) -> None:
    if self._bar_lift is not None:
      self._bar_lift.next()   # type:ignore[no-untyped-call]

  def _core_asm_lift_done(self) -> None:
    if self._bar_lift:
      self._bar_lift.finish(end='\r')

  def _core_asm_disasm_begin(self) -> None:
    if ui.is_tty():
      from progressbar import ProgressBar, RotatingMarker
      self._bar_disasm = ProgressBar(widgets=[
        ui.bullet('info'),
        'disassemble: disassembling... ',
        RotatingMarker()   # type:ignore[no-untyped-call]
      ])
    else:
      ui.info('disassemble: disassembling... ')
      self._bar_disasm = None

  def _core_asm_disasm_update(self) -> None:
    if self._bar_disasm is not None:
      self._bar_disasm.next()   # type:ignore[no-untyped-call]

  def _core_asm_disasm_done(self) -> None:
    if self._bar_disasm is not None:
      ui.info('disassemble: disassembling... done.', ow=True)
    else:
      ui.info('disassemble: done')

  def _core_asm_asm_begin(self) -> None:
    if ui.is_tty():
      from progressbar import ProgressBar, RotatingMarker
      self._bar_asm = ProgressBar(widgets=[
        ui.bullet('info'),
        'assemble: assembling... ',
        RotatingMarker()   # type:ignore[no-untyped-call]
      ])
    else:
      ui.info('assemble: assembling... ')
      self._bar_asm = None

  def _core_asm_asm_update(self) -> None:
    if self._bar_asm is not None:
      self._bar_asm.next()   # type:ignore[no-untyped-call]

  def _core_asm_asm_done(self) -> None:
    if self._bar_asm is not None:
      ui.info('assemble: assembling... done.', ow=True)
    else:
      ui.info('assemble: done')

  def _core_context_disasm_done(self) -> None:
    ui.info('analyze: disassembling... done.', ow=True)

  def _core_analysis_smali_begin(self, total: int) -> None:
    if ui.is_tty():
      from progressbar import ProgressBar, Percentage, GranularBar, SimpleProgress, ETA
      self._bar_analysis = ProgressBar(
        max_value=total,
        widgets=[
          ui.bullet('info'),
          'analyze: analyzing... ',
          Percentage(), ' ',  # type:ignore[no-untyped-call]
          GranularBar(), ' ',  # type:ignore[no-untyped-call]
          SimpleProgress(format='%(value_s)s/%(max_value_s)s'), ' ',   # type:ignore[no-untyped-call]
          '(', ETA(), ')'   # type:ignore[no-untyped-call]
        ]
      )
    else:
      self._bar_analysis = None

  def _core_analysis_smali_analyzing(self, nr: int) -> None:
    if self._bar_analysis is not None:
      self._bar_analysis.update(nr)
    else:
      if (nr % 128) == 0:
        ui.info(f"analyze: analyzing ... {nr} classes")

  def _core_analysis_smali_analyzed(self) -> None:
    if self._bar_analysis is not None:
      self._bar_analysis.finish()

  def _core_analysis_smali_summary(self, ops: Optional[int] = None, classes: Optional[int] = None, methods: Optional[int] = None) -> None:
    if self._bar_analysis:
      o: str = ''
      tail: str = ' '*20
      if ops is None:
        o += f'ops ... {tail}'
      else:
        o += f'{ops} ops, '
        if classes is None:
          o += f'classes ... {tail}'
        else:
          o += f'{classes} classes, '
          if methods is None:
            o += f'methods ... {tail}'
          else:
            o += f'{methods} methods.'
      ui.info(f"analyze: got {o}", nl=(methods is not None), ow=True)
    else:
      if methods is not None:
        ui.info(f"analyze: methods: {methods}")
      elif classes is not None:
        ui.info(f"analyze: classes: {classes}")
      elif ops is not None:
        ui.info(f"analyze: ops: {ops}")

  def _core_analysis_smali_finalizing(self) -> None:
    ui.info("analyze: finalizing")

  def _core_analysis_smali_done(self, t: float) -> None:
    ui.info(f"analyze: done ({t:.02f} sec)")

  def _core_analysis_nat_begin(self) -> None:
    ui.info('analyze_nat: analyzing...', nl=False)

  def _core_analysis_nat_analyzing(self) -> None:
    ui.info('analyze_nat: calls ...{tail}'.format(tail=' '*20), ow=True, nl=False)

  def _core_analysis_nat_summary(self, calls: Optional[int] = None) -> None:
    ui.info(f'analyze_nat: got {calls} calls', ow=True)

  def _core_analysis_nat_done(self, t: float) -> None:
    ui.info(f"analyze_nat: done ({t:.02f} sec)")

  def _core_analysis_done(self, t: float) -> None:
    ui.info(f"analyze: done ({t:.02f} sec)")

class FileTransferProgressReporter:
  _bar: Optional[ProgressBar]
  def __init__(self, desc: str) -> None:
    self._done = False
    self._desc = desc

  @contextmanager
  def scoped(self) -> Iterator[FileTransferProgressReporter]:
    try:
      self.start()
      yield self
    finally:
      if self._bar is not None and not self._done:
        self._bar.finish()

  def using_bar(self) -> bool:
    return self._bar is not None

  def start(self) -> None:
    if ui.is_tty():
      from progressbar import ProgressBar, Counter, RotatingMarker
      self._bar = ProgressBar(widgets=[
        ui.bullet('info'),
        f'{self._desc}... ',
        Counter(), ' ',   # type:ignore[no-untyped-call]
        RotatingMarker()   # type:ignore[no-untyped-call]
      ])
      self._bar.start()
    else:
      ui.info(self._desc)
      self._bar = None

  def update(self, nr: int) -> None:
    if self._bar is not None:
      self._bar.update(nr)
    else:
      ui.info(f' .. {nr} files')

  def done(self) -> None:
    if self._bar is not None:
      self._bar.finish(end='\r')
      ui.info(f'{self._desc}... done.' + (' '*16), ow=True)
    else:
      ui.info(f'{self._desc}... done.')
    self._done = True

class ScanProgressReporter:
  def __init__(self) -> None:
    from trueseeing.core.report import ConsoleNoter
    self._CN = ConsoleNoter()

  @contextmanager
  def scoped(self) -> Iterator[None]:
    submap = {
      'issue':self._issue,
    }
    try:
      for k, v in submap.items():
        pub.subscribe(v, k)
      yield None
    finally:
      for k, v in submap.items():
        pub.unsubscribe(v, k)
      pass

  def _issue(self, issue: Issue) -> None:
    self._CN.note(issue)

class AndroidInstallProgressReporter:
  _bar: Optional[ProgressBar] = None

  @contextmanager
  def scoped(self) -> Iterator[None]:
    submap = {
      'progress.android.adb.begin':self._begin,
      'progress.android.adb.update':self._update,
      'progress.android.adb.done':self._done,
    }
    try:
      for k, v in submap.items():
        pub.subscribe(v, k)
      yield None
    finally:
      for k, v in submap.items():
        pub.unsubscribe(v, k)
      pass

  def _begin(self, what: str) -> None:
    if ui.is_tty():
      from progressbar import ProgressBar, RotatingMarker
      self._bar = ProgressBar(widgets=[
        ui.bullet('info'),
        what + ' ',
        RotatingMarker()   # type:ignore[no-untyped-call]
      ])
    else:
      self._bar = None

  def _update(self) -> None:
    if self._bar is not None:
      self._bar.next()   # type:ignore[no-untyped-call]

  def _done(self) -> None:
    if self._bar:
      self._bar.finish(end='\r')

class OpFormatter:
  def __init__(self, q: APKQuery, indent: int = 4) -> None:
    self._q = q
    self._indent = indent

  def format(self, ops: Iterable[Op]) -> Iterator[Tuple[bool, str]]:
    focus: Optional[str] = None
    for op in ops:
      qn = self._q.qualname_of(op.addr)
      if qn is None:
        qn = self._q.class_name_of(op.addr)
      if qn != focus:
        yield True, f'{qn}:'
        focus = qn
      yield False, '{ind}{op}'.format(
        ind=' '*self._indent,
        op='{id:08x}{sep}{l}'.format(sep=' '*4, id=op.addr, l=op.l.lstrip())
      )

class OpLister:
  def __init__(self, formatter: OpFormatter) -> None:
    self._formatter = formatter

  def list_tagged(self, ops: Iterable[Op]) -> None:
    for is_header, line in self._formatter.format(ops):
      if is_header:
        ui.info(ui.colored(line, color='green'))
      else:
        ui.info(line)

  def list_untagged(self, ops: Iterable[Op]) -> None:
    for is_header, line in self._formatter.format(ops):
      if not is_header:
        ui.info(line)


ui = UI()



from __future__ import annotations
from typing import TYPE_CHECKING

import math
import re
from trueseeing.api import Signature
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Dict, Optional, Mapping, Any, AnyStr, Set
  from trueseeing.api import SignatureMap, SignatureHelper, ConfigMap
  from trueseeing.core.ios.model import Call
  from trueseeing.core.ios.db import IPAQuery
  from trueseeing.core.ios.context import IPAContext

class IOSDetector(Signature):
  _cvss_info = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'

  def __init__(self, helper: SignatureHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return IOSDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {
      'ios-nat-api':dict(e=self._detect_api, d='[iOS] Detects API call'),
      'ios-nat-urls':dict(e=self._detect_url, d='[iOS] Detects URL etc.'),
      'ios-detect-dyncode':dict(e=self._detect_dyncode, d='[iOS] Detects dynamic code exec attempt'),
      'ios-detect-syscall':dict(e=self._detect_inconsistent_syscall, d='[iOS] Detects syscalls looks inconsistent'),
      'ios-detect-reflection':dict(e=self._detect_reflection, d='[iOS] Detects possible reflections'),
      'ios-detect-jb':dict(e=self._detect_jb, d='[iOS] Detects possible JB probes'),
      'ios-detect-debug':dict(e=self._detect_debug, d='[iOS] Detects possible debug probes'),
      'ios-detect-privacy':dict(e=self._detect_privacy, d='[iOS] Detects privacy concerns'),
      'ios-detect-obfus':dict(e=self._detect_obfuscation, d='[iOS] Detects obfuscated functions'),
      'ios-detect-assert':dict(e=self._detect_assert, d='[iOS] Detects assertions'),
      'ios-detect-logging':dict(e=self._detect_log, d='[iOS] Detects log'),
      'ios-detect-lib-needed':dict(e=self._detect_needed_libs, d='[iOS] Detects needed libraries'),
      'ios-detect-motion':dict(e=self._detect_motion, d='[iOS] Detects use of motion/gyro etc.'),
      'ios-detect-urlscheme':dict(e=self._detect_urlscheme, d='[iOS] Detects recognized URL schemes'),
      'ios-detect-ats':dict(e=self._detect_ats, d='[iOS] Detects ATS status'),
      'ios-detect-permission':dict(e=self._detect_permission, d='[iOS] Detects protected resource accesses'),
      'ios-detect-req':dict(e=self._detect_device_req, d='[iOS] Detected device requirements'),
      'ios-detect-device':dict(e=self._detect_device_info, d='[iOS] Detects device info probes'),
      'ios-detect-ents':dict(e=self._detect_entitlements, d='[iOS] Detects entitlements'),
      'ios-detect-copyrights':dict(e=self._detect_copyrights, d='[iOS] Detects copyright banners'),
      'ios-detect-crypto-xor':dict(e=self._detect_crypto_xor, d='[iOS] Detects Vernum cipher usage with static keys'),
      'ios-detect-libs':dict(e=self._detect_libs, d='[iOS] Detects statically-linked libs'),
    }

  def get_configs(self) -> ConfigMap:
    return dict()

  def _get_ipa_context(self) -> IPAContext:
    return self._helper.get_context().require_type('ipa') # type:ignore[return-value]

  def _format_aff0(self, c: Call) -> str:
    return self._format_aff0_manual(c['path'], c['sect'], c['offs'])

  def _format_aff0_match(self, n: str, m: re.Match[AnyStr]) -> str:
    return self._format_aff0_manual(n, '', m.start())

  def _format_aff0_manual(self, n: str, s: str, o: int) -> str:
    return '{} ({}+{:08x})'.format(n, s, o)

  async def _detect_api(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      from trueseeing.core.ios.swift import SwiftDemangler
      async with SwiftDemangler.get(simplify=True).scoped() as dem:
        for c in q.calls():
          priv, target, swift = c['priv'], c['target'], c['swift']
          if swift:
            target = await dem.resolve(target)
          self._helper.raise_issue(self._helper.build_issue(
            sigid='ios-nat-api',
            cvss=self._cvss_info,
            title='detected {} call'.format('private' if priv else 'API'),
            info0=target,
            aff0=self._format_aff0(c),
          ))

  async def _detect_url(self) -> None:
    from trueseeing.core.analyze import analyze_url_in
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for d in analyze_url_in(q.file_enum('disasm/%', neg=True)):
        tentative = False
        if '...' in d['v']:
          ui.warn('truncated value found; disassemble again with wider fields', onetime=True)
          tentative = True
        self._helper.raise_issue(self._helper.build_issue(
          sigid='ios-nat-urls',
          cvss=self._cvss_info,
          title='detected {}'.format(d['typ']),
          cfd='tentative' if tentative else 'firm',
          info0=d['v'],
          aff0=d['fn'],
        ))

  async def _detect_dyncode(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls():
        if '_text' in c['sect']:
          if not c['priv'] and '_dlopen' in c['target']:
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-dynload',
              cvss=self._cvss_info,
              title='dynamically loading code',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

          if re.search(r'VmStack|vm_stack|vm_err|StackPool|stack_pool|FunctionCopy|push_[if]64\(', c['target']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-dynload',
              cvss=self._cvss_info,
              title='detected VM-like impl',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_inconsistent_syscall(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls(api=True):
        if '_text' in c['sect']:
          if any((x in c['target']) for x in ['_fork', '_vfork', '_syscall']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-syscall',
              cvss=self._cvss_info,
              title='inconsistent syscall',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_reflection(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls(api=True):
        if '_text' in c['sect']:
          if any((x in c['target']) for x in ['NSInvocation', '_NSClassFromString']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-reflection',
              cvss=self._cvss_info,
              title='use of reflection',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))
          elif '_class_add' in c['target']:
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-reflection',
              cvss=self._cvss_info,
              title='monkeypatching of class',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_jb(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls(api=True):
        if '_text' in c['sect']:
          if any((x in c['target']) for x in ['Jail', 'jailb']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-jb',
              cvss=self._cvss_info,
              title='possible JB probe',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_debug(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls(api=True):
        if '_text' in c['sect']:
          if any((x in c['target']) for x in ['Debuggi', 'debuggi']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-debug',
              cvss=self._cvss_info,
              title='possible debug probe',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_privacy(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls():
        mentioned = False
        if '_text' in c['sect']:
          if not c['priv']:
            if ' uniqueDeviceIdentifer]' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('UDID'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            if ' uniqueGlobalDeviceIdentifer]' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('UGDID'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif ' identifierForVe' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('IDFV'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif ' resettable' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('resettable device id'),
                cfd='tentative',
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif re.search(r' currentCarrier| carrierName\]', c['target']):
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('carrier'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif ' mobileCountryCode' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('MCC'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif ' mobileNetworkCode' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('MNC'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif ' freeDiskspace' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('device info (diskspace)'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif ' currencyCode' in c['target']:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('user preference (currency)'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            if re.search(r'\[GMSCoordinateBounds (isValid|(north|south)(West|East))\]|_CLLocation', c['target']):
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format('location'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif re.search(r'pasteboard.*?change', c['target'], flags=re.IGNORECASE):
              mentioned = True
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: watching pasteboard',
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif re.search(r'(data|value)forpasteboard', c['target'], flags=re.IGNORECASE):
              mentioned = True
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: reading pasteboard',
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
            elif re.search(r'setdata:.*?forpasteboardtype', c['target'], flags=re.IGNORECASE):
              mentioned = True
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: writing pasteboard',
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))

          if 'Pasteboard' in c['target']:
            if not mentioned:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: referring pasteboard',
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))
          elif '[ABKDevice ' in c['target']:
            m = re.search(r'\[ABKDevice (.*?)\]', c['target'])
            parts = m.group(1) if m else 'unknown'
            if not parts.startswith('set'):
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-privacy',
                cvss=self._cvss_info,
                title='privacy concern: getting {}'.format(f'device info ({parts} [ABKDevice])'),
                info0=c['target'],
                aff0=self._format_aff0(c),
              ))

  async def _detect_obfuscation(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls():
        if '_text' in c['sect']:
          m = re.search(r' _[A-Z]+([a-z]{3,5}_[a-z]{2,4}(?:_[a-z]{2,6})?)\(', c['target'])
          if m:
            ent = self._entropy_of(m.group(1))
            if ent < 2.8 and not re.search(r' _BDP(mpi|rsa)', c['target']):
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-obfus',
                cvss=self._cvss_info,
                title='possible obfuscated function',
                cfd='tentative',
                info0=c['target'],
                info1=f'{ent:.04f}',
                aff0=self._format_aff0(c),
              ))

  async def _detect_assert(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls(api=True):
        if '_text' in c['sect']:
          if '_assertionFailure_' in c['target']:
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-assert',
              cvss=self._cvss_info,
              title='possible live assertion check',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_log(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls(api=True):
        if '_text' in c['sect']:
          if '_NSLog' in c['target']:
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-logging',
              cvss=self._cvss_info,
              title='detected logging',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_needed_libs(self) -> None:
    from trueseeing.core.ios.analyze import analyze_lib_needs_in
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for d in analyze_lib_needs_in(q.file_enum('^disasm/|/_CodeSignature/', neg=True, regex=True)):
        self._helper.raise_issue(self._helper.build_issue(
          sigid='ios-detect-lib-needed',
          cvss=self._cvss_info,
          title='detected library reference',
          cfd='firm',
          info0=d['v'],
          aff0=d['fn'],
        ))

  async def _detect_motion(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls(api=True):
        if '_text' in c['sect']:
          if re.search(r'startGyro| (gyroUpdateInterval|rotationRate)\]', c['target']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-motion',
              cvss=self._cvss_info,
              title='watching gyroscope',
              cfd='firm',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))
          if re.search(r'startAccelero| (accelerometerUpdateInterval|(a|userA)cceleration)\]', c['target']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-motion',
              cvss=self._cvss_info,
              title='watching accelerometer',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))
          if re.search(r'startDeviceMotion| (deviceMotionUpdateInterval|gravity)\]', c['target']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-motion',
              cvss=self._cvss_info,
              title='watching motion',
              cfd='tentative',
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_urlscheme(self) -> None:
    q: IPAQuery
    from plistlib import loads
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for n, b in q.file_enum('Payload/.*?.app/Info.plist', regex=True):
        dom = loads(b)
        if 'LSApplicationQueriesSchemes' in dom:
          v = dom['LSApplicationQueriesSchemes']
          for scheme in v:
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-urlscheme',
              cvss=self._cvss_info,
              title='probing URL scheme',
              cfd='firm',
              info0=scheme,
              aff0=n,
            ))
        if 'CFBundleURLTypes' in dom:
          v = dom['CFBundleURLTypes']
          for d in v:
            for scheme in d.get('CFBundleURLSchemes', []):
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-urlscheme',
                cvss=self._cvss_info,
                title='handling URL scheme',
                cfd='firm',
                info0=scheme,
                info1=d.get('CFBundleURLName', '(unknown name)'),
                info2=d.get('CFBundleTypeRole', '(unknown role)'),
                aff0=n,
              ))

  async def _detect_ats(self) -> None:
    q: IPAQuery
    from plistlib import loads
    context = self._get_ipa_context()
    cvss0 = 'CVSS:3.0/AV:A/AC:L/PR:N/UI:N/S:U/C:L/I:L/A:N/'
    cvss1 = 'CVSS:3.0/AV:A/AC:H/PR:N/UI:N/S:U/C:L/I:L/A:N/'
    with context.store().query().scoped() as q:
      for n, b in q.file_enum('Payload/.*?.app/Info.plist', regex=True):
        dom = loads(b)
        if 'NSAppTransportSecurity' in dom:
          v = dom['NSAppTransportSecurity']
          if v.get('NSAllowsArbitraryLoads'):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-ats',
              cvss=cvss0,
              title='disabing ATS',
              cfd='firm',
              aff0=n,
            ))
          if v.get('NSAllowsArbitraryLoadsInWebContent'):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-ats',
              cvss=cvss1,
              title='disabling ATS for web content',
              cfd='firm',
              aff0=n,
            ))
          if v.get('NSAllowsArbitraryLoadsForMedia'):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-ats',
              cvss=cvss1,
              title='disabling ATS for media',
              cfd='firm',
              aff0=n,
            ))
          if v.get('NSAllowsLocalNetworking'):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-ats',
              cvss=cvss1,
              title='use of local networking',
              cfd='firm',
              aff0=n,
            ))
          if v.get('NSAllowsLocalNetworking'):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-ats',
              cvss=cvss1,
              title='use of local networking',
              cfd='firm',
              aff0=n,
            ))
          if v.get('NSExceptionDomains'):
            for name, desc in v['NSExceptionDomains'].items():
              include_subdomains = desc.get('NSIncludeSubdomains', False)
              if desc.get('NSExceptionAllowsInsecureHTTPLoads'):
                self._helper.raise_issue(self._helper.build_issue(
                  sigid='ios-detect-ats',
                  cvss=cvss1,
                  title='partially disabing ATS',
                  cfd='firm',
                  info0='{}{}'.format('*.' if include_subdomains else '', name),
                  aff0=n,
                ))
              else:
                self._helper.raise_issue(self._helper.build_issue(
                  sigid='ios-detect-ats',
                  cvss=self._cvss_info,
                  title='partially enabling ATS',
                  cfd='firm',
                  info0='{}{}'.format('*.' if include_subdomains else '', name),
                  aff0=n,
                ))
              if desc.get('NSExceptionMinimumTLSVersion', 'TLSv1.3').lower() != 'tlsv1.3':
                self._helper.raise_issue(self._helper.build_issue(
                  sigid='ios-detect-ats',
                  cvss=cvss1,
                  title='partial use of lower TLS version',
                  cfd='firm',
                  info0='{}{}'.format('*.' if include_subdomains else '', name),
                  info1=desc['NSExceptionMinimumTLSVersion'],
                  aff0=n,
                ))
              if not desc.get('NSExceptionRequiresForwardSecrecy'):
                self._helper.raise_issue(self._helper.build_issue(
                  sigid='ios-detect-ats',
                  cvss=cvss0,
                  title='partially allowing use of non-PFS ciphers',
                  cfd='firm',
                  info0='{}{}'.format('*.' if include_subdomains else '', name),
                  aff0=n,
                ))
              if not desc.get('NSRequiresCertificateTransparency'):
                self._helper.raise_issue(self._helper.build_issue(
                  sigid='ios-detect-ats',
                  cvss=self._cvss_info,
                  title='attempts to partially disable CT',
                  cfd='firm',
                  info0='{}{}'.format('*.' if include_subdomains else '', name),
                  aff0=n,
                ))
          if v.get('NSPinnedDomains'):
            from base64 import b64decode
            from binascii import hexlify
            for name, desc in v['NSPinnedDomains'].items():
              include_subdomains = desc.get('NSIncludeSubdomains', False)
              if desc.get('NSPinnedLeafIdentities'):
                idents = desc['NSPinnedLeafIdentities']
                for ident in idents:
                  for algo, h in ident.items():
                    m = re.fullmatch(r'SPKI-(.*?)-BASE64', algo)
                    if m:
                      algo = m.group(1)
                      h = hexlify(b64decode(h)).decode()
                    self._helper.raise_issue(self._helper.build_issue(
                      sigid='ios-detect-ats',
                      cvss=self._cvss_info,
                      title='detected TLS certificate pinning',
                      cfd='firm',
                      info0='{}{}'.format('*.' if include_subdomains else '', name),
                      info1='{} ({})'.format(h, algo),
                      aff0=n,
                    ))

  async def _detect_permission(self) -> None:
    q: IPAQuery
    from plistlib import loads
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for n, b in q.file_enum('Payload/.*?.app/Info.plist', regex=True):
        dom = loads(b)

        m = dict(
          NSBluetoothAlwaysUsageDescription='persistent Bluetooth access',
          NSBluetoothPeripheralUsageDescription='Bluetooth peripheral access',
          NSCalendarsFullAccessUsageDescription='calendar read/write access',
          NSCalendarsWriteOnlyAccessUsageDescription='calendar write access',
          NSRemindersFullAccessUsageDescription='reminder read/write access',
          NSCameraUsageDescription='camera access',
          NSMicrophoneUsageDescription='microphone access',
          NSContactsUsageDescription='contact read access',
          NSFaceIDUsageDescription='FaceID access',
          NSGKFriendListUsageDescription='GameCenter friend list access',
          NSHealthClinicalHealthRecordsShareUsageDescription='clinical record read access',
          NSHealthShareUsageDescription='HealthKit sample read access',
          NSHealthUpdateUsageDescription='HealthKit sample write access',
          NSHomeKitUsageDescription='HomeKit config read access',
          NSLocationAlwaysAndWhenInUseUsageDescription='persistent location access',
          NSLocationUsageDescription='location access',
          NSLocationWhenInUseUsageDescription='foreground location access',
          NSLocationTemporaryUsageDescriptionDictionary='temporary location access',
          NSLocationAlwaysUsageDescription='persistent location access',
          NSAppleMusicUsageDescription='media library access',
          NSMotionUsageDescription='motion read access',
          NSFallDetectionUsageDescription='fall detection read access',
          NSLocalNetworkUsageDescription='local network access',
          NSNearbyInteractionUsageDescription='nearby device interaction access',
          NSNearbyInteractionAllowOnceUsageDescription='temporary nearby device interaction access',
          NFCReaderUsageDescription='NFC read access',
          NSPhotoLibraryAddUsageDescription='photo library append access',
          NSPhotoLibraryUsageDescription='photo library read/write access',
          NSUserTrackingUsageDescription='device tracking access',
          NSSensorKitUsageDescription='sensor read access',
          NSSiriUsageDescription='Siri integration access',
          NSSpeechRecognitionUsageDescription='speech recognition access',
          NSVideoSubscriberAccountUsageDescription='TV provider account access',
          NSWorldSensingUsageDescription='world-sensing data access',
          NSHandsTrackingUsageDescription='hand tracking data access',
          NSIdentityUsageDescription='Wallet identity access',
          NSCalendarsUsageDescription='calendar read/write access',
          NSRemindersUsageDescription='reminder read/write access',
        )

        for k, res in m.items():
          v = dom.get(k)
          if v:
            if isinstance(v, dict):
              for vk, vv in v.items():
                self._helper.raise_issue(self._helper.build_issue(
                  sigid='ios-detect-permission',
                  cvss=self._cvss_info,
                  title=f'declaring {res}',
                  cfd='firm',
                  info0='{} ({})'.format(vv, vk),
                  aff0=n,
                ))
            else:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-permission',
                cvss=self._cvss_info,
                title=f'declaring {res}',
                cfd='firm',
                info0=v,
                aff0=n,
              ))

  async def _detect_device_req(self) -> None:
    q: IPAQuery
    from plistlib import loads
    context = self._get_ipa_context()
    cvss0 = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:L/I:L/A:N/'
    with context.store().query().scoped() as q:
      for n, b in q.file_enum('Payload/.*?.app/Info.plist', regex=True):
        dom = loads(b)
        d = dom.get('UISupportedDevices')
        if d:
          for k in d:
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-device',
              cvss=self._cvss_info,
              title='detected required device model',
              cfd='firm',
              info0=k,
              aff0=n,
            ))
        d = dom.get('UIDeviceFamily')
        if d:
          for k in d:
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-device',
              cvss=self._cvss_info,
              title='detected targetted device family',
              cfd='firm',
              info0={1:'iPhone/iPod Touch', 2:'iPad/Mac Catalyst (for iPad)', 3:'Apple TV', 4:'Apple Watch', 6:'Mac Catalyst (for Mac)', 7:'Apple Vision'}.get(k, '(unknown)'),
              aff0=n,
            ))
        d = dom.get('LSRequiresIPhoneOS')
        if not d:
          self._helper.raise_issue(self._helper.build_issue(
            sigid='ios-detect-device',
            cvss=cvss0,
            title='insufficient device integrity check: iOS is not required',
            cfd='tentative',
            aff0=n,
          ))
        d = dom.get('MinimumOSVersion')
        if d:
          self._helper.raise_issue(self._helper.build_issue(
            sigid='ios-detect-device',
            cvss=self._cvss_info,
            title='detected minimum OS version',
            cfd='firm',
            info0=d,
            aff0=n,
          ))
        d = dom.get('UIRequiredDeviceCapabilities')
        if d:
          for k in ['arm64', 'armv7']:
            if k in d:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-device',
                cvss=self._cvss_info,
                title='detected required architecture',
                cfd='firm',
                info0=k,
                aff0=n,
              ))

  async def _detect_device_info(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for c in q.calls():
        if '_text' in c['sect']:
          if re.search(r' osName]', c['target']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-device',
              cvss=self._cvss_info,
              title='getting device {}'.format('OS name'),
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))
          if re.search(r' osVersion\]|OSVersion', c['target']) and not re.search(r'support|minim|atleast', c['target']):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-device',
              cvss=self._cvss_info,
              title='getting device {}'.format('OS version'),
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))
          if re.search(r' (device)?model\]', c['target'], re.IGNORECASE):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-device',
              cvss=self._cvss_info,
              title='getting device {}'.format('model number'),
              info0=c['target'],
              aff0=self._format_aff0(c),
              cfd='firm' if 'device' in c['target'] else 'tentative',
            ))
          if re.search(r'sysctl', c['target'], re.IGNORECASE):
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-device',
              cvss=self._cvss_info,
              title='getting device {}'.format('kernel tunables'),
              info0=c['target'],
              aff0=self._format_aff0(c),
            ))

  async def _detect_entitlements(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for n, b in q.file_enum(r'Payload/(.*?).app/\1', regex=True):
        ents = self._get_entitlements(b)
        if ents:
          for k, v in ents.items():
            self._helper.raise_issue(self._helper.build_issue(
              sigid='ios-detect-ents',
              cvss=self._cvss_info,
              title='detected entitlement',
              cfd='firm',
              info0=k,
              info1=repr(v),
              aff0=n,
            ))

  async def _detect_copyrights(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for n, b in q.file_enum('Payload/%'):
        for m in re.finditer(rb'(?:Copyright )?\(C\) [\x20-\xff]+', b, re.IGNORECASE):
          banner = m.group(0)
          if any((x in banner) for x in [b'You must retain', b'allow any third party to access', b' -> ']):
            continue
          m0 = re.match(rb'(.*)(?:<[a-z]+ ?/>|</[a-z]+>|\*/)$', banner)
          if m0:
            banner = m0.group(1)
          self._helper.raise_issue(self._helper.build_issue(
            sigid='ios-detect-ents',
            cvss=self._cvss_info,
            title='detected copyright banner',
            cfd='firm',
            info0=banner.decode('latin1'),
            aff0=self._format_aff0_match(n, m),
          ))

  async def _detect_crypto_xor(self) -> None:
    from trueseeing.core.ios.analyze import get_origin
    q: IPAQuery
    context = self._get_ipa_context()
    with context.store().query().scoped() as q:
      for n, b in q.file_enum('disasm/%.s'):
        for m in re.finditer(rb'^.*? eor .*#(0x[0-9a-f]+)', b, re.MULTILINE):
          mask = m.group(1)
          if re.search(rb'^0x[124837cf]$|^0x[12483cf]0$|^0x[137bde2c]f$|00$|ff$|ffff|0000', mask):
            continue
          o = get_origin(n, m.group(0))
          self._helper.raise_issue(self._helper.build_issue(
            sigid='ios-detect-crypto-xor',
            cvss=self._cvss_info,
            title='detected non-random XOR cipher',
            cfd='tentative',
            info0='0x{:02x}'.format(int(mask, 16)),
            aff0=self._format_aff0_manual(o['fn'], o['sect'], o['offs']),
          ))

  async def _detect_libs(self) -> None:
    q: IPAQuery
    context = self._get_ipa_context()
    seen: Dict[str, Dict[str, Set[str]]] = dict(det=dict(), ref=dict())
    with context.store().query().scoped() as q:
      for c in q.calls():
        if '_text' in c['sect']:
          m = re.search(r'[\[ ]([A-Z]{2,})[A-Z][a-z]', c['target'])
          if m:
            prefix = m.group(1)
            path = c['path']
            ref = not c['priv']
            sk = 'ref' if ref else 'det'
            try:
              lseen = seen[sk][path]
            except KeyError:
              lseen = set()
              seen[sk][path] = lseen
            if prefix not in lseen:
              self._helper.raise_issue(self._helper.build_issue(
                sigid='ios-detect-libs',
                cvss=self._cvss_info,
                title='detected library{}'.format(' ref' if ref else ''),
                info0=prefix,
                aff0=self._format_aff0(c),
              ))
              lseen.add(prefix)

  def _get_entitlements(self, app: bytes) -> Optional[Mapping[str, Any]]:
    from plistlib import loads
    m = re.search(rb'<\?xml [^\x00-\x08\x0b-\x1f\x80-\xff]+?<plist [^\x00-\x08\x0b-\x1f\x80-\xff]+?application-identifier[^\x00-\x08\x0b-\x1f\x80-\xff]+?</plist>', app, re.DOTALL)
    if m:
      return loads(m.group(0))  # type:ignore[no-any-return]
    else:
      return None

  @classmethod
  def _entropy_of(cls, string: str) -> float:
    o = 0.0
    m: Dict[str, int] = dict()
    for c in string:
      m[c] = m.get(c, 0) + 1
    for cnt in m.values():
      freq = float(cnt) / len(string)
      o -= freq * (math.log(freq) / math.log(2))
    return o

  @classmethod
  def _assumed_randomness_of(cls, string: str) -> float:
    try:
      return cls._entropy_of(string) / float(math.log(len(string)) / math.log(2))
    except ValueError:
      return 0



from __future__ import annotations
from typing import TYPE_CHECKING

from functools import cache
import os

if TYPE_CHECKING:
  from typing import Optional

@cache
def get_home_dir() -> str:
  return os.environ.get('TS2_HOME', os.path.join(os.environ['HOME'], '.trueseeing2'))

@cache
def get_rc_path() -> str:
  return os.path.join(get_home_dir(), 'rc')

@cache
def get_cache_dir() -> str:
  return get_cache_dir_v2()

@cache
def get_cache_dir_v0() -> str:
  return get_home_dir()

@cache
def get_cache_dir_v1(target: str) -> str:
  return os.environ.get('TS2_CACHEDIR', os.path.dirname(target))

@cache
def get_cache_dir_v2() -> str:
  return os.environ.get('TS2_CACHEDIR', os.path.join(get_home_dir(), 'cache'))

@cache
def get_adb_host() -> Optional[str]:
  return os.environ.get('TS2_ADB_HOST', ('tcp:host.docker.internal:5037' if is_in_container() else None))

@cache
def is_in_container() -> bool:
  return 'TS2_IN_DOCKER' in os.environ

@cache
def get_shell() -> str:
  return os.environ.get('SHELL', '/bin/sh')

@cache
def get_extension_dir() -> str:
  return os.environ.get('TS2_EXTDIR', os.path.join(get_home_dir(), 'extensions'))

@cache
def get_extension_dir_v0() -> str:
  return get_home_dir()

@cache
def get_extension_package_prefix() -> str:
  return 'trueseeing_ext0_'

@cache
def get_device_frida_dir(package_name: str) -> str:
  return '/data/local/tmp/ts2/{pkg}/frida'.format(pkg=package_name)

@cache
def get_swift_demangler_url() -> str:
  return os.environ.get('TS2_SWIFT_DEMANGLER_URL', 'http://127.0.0.1:8000')

@cache
def get_cache_schema_id() -> int:
  return 0x54f6d672  # FIXME: remember to randomize this whenever incompatible changes occur on cache file structure, or DB schema



from __future__ import annotations
from typing import TYPE_CHECKING

from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Any, Optional, Dict, Mapping, List, Iterator, Tuple
  from trueseeing.api import CommandHelper, Command, CommandMap, OptionMap
  from trueseeing.core.android.context import APKContext
  from trueseeing.core.android.model import XAPKManifest

class EngageCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return EngageCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      'xtq':dict(e=self._engage_tamper_discard, n='xtq', d='engage: discard changes', t={'apk'}),
      'xtx':dict(e=self._engage_tamper_apply, n='xtx[!]', d='engage: apply and rebuild apk', t={'apk'}),
      'xtx!':dict(e=self._engage_tamper_apply, t={'apk'}),
      'xtf':dict(e=self._engage_tamper_inject_frida, n='xtf[!] [config]', d='engage; inject frida gadget', t={'apk'}),
      'xtf!':dict(e=self._engage_tamper_inject_frida, t={'apk'}),
      'xtfs':dict(e=self._engage_tamper_inject_frida_scriptdir, n='xtfs[!] [path]', d='engage; inject frida gadget in script dir mode', t={'apk'}),
      'xtfs!':dict(e=self._engage_tamper_inject_frida_scriptdir, t={'apk'}),
      'xtn':dict(e=self._engage_tamper_disable_pinning, n='xtn', d='engage: patch NSC to disable SSL/TLS pinning', t={'apk'}),
      'xtd':dict(e=self._engage_tamper_enable_debug, n='xtd', d='engage: make debuggable', t={'apk'}),
      'xtb':dict(e=self._engage_tamper_enable_backup, n='xtb', d='engage: make backupable', t={'apk'}),
      'xtt':dict(e=self._engage_tamper_patch_target_api_level, n='xtt[!] <api level>', d='engage: patch target api level', t={'apk'}),
      'xtt!':dict(e=self._engage_tamper_patch_target_api_level, t={'apk'}),
      'xco':dict(e=self._engage_device_copyout, n='xco[!] package [data.tar]', d='engage: copy-out package data', t={'apk'}),
      'xco!':dict(e=self._engage_device_copyout, t={'apk'}),
      'xci':dict(e=self._engage_device_copyin, n='xci[!] package [data.tar]', d='engage: copy-in package data', t={'apk'}),
      'xci!':dict(e=self._engage_device_copyin, t={'apk'}),
      'xpd':dict(e=self._engage_deploy_package, n='xpd[!]', d='engage: deploy target package', t={'apk'}),
      'xpd!':dict(e=self._engage_deploy_package, t={'apk'}),
      'xpu':dict(e=self._engage_undeploy_package, n='xpu', d='engage: remove target package', t={'apk'}),
      'xz':dict(e=self._engage_fuzz_intent, n='xz[!] "am-cmdline-template" [output.txt]', d='engage: fuzz intent', t={'apk'}),
      'xz!':dict(e=self._engage_fuzz_intent, t={'apk'}),
      'xzr':dict(e=self._engage_fuzz_command, n='xzr[!] "cmdline-template" [output.txt]', d='engage: fuzz cmdline', t={'apk'}),
      'xzr!':dict(e=self._engage_fuzz_command, t={'apk'}),
      'xg':dict(e=self._engage_grab_package, n='xg[!] package [output.apk]', d='engage: grab package', t={'apk'}),
      'xg!':dict(e=self._engage_grab_package, t={'apk'}),
      'xs':dict(e=self._engage_frida_start_server, n='xs[!] [config]', d='engage: start frida-server (!: force)', t={'apk'}),
      'xs!':dict(e=self._engage_frida_start_server, t={'apk'}),
      'xk':dict(e=self._engage_frida_kill_server, n='xk', d='engage: kill frida server', t={'apk'}),
    }

  def get_options(self) -> OptionMap:
    return {
      'vers':dict(n='vers=X.Y.Z', d='specify frida-gadget version to use [xf,xfs]', t={'apk'}),
      'w':dict(n='wNAME=FN', d='wordlist, use as {NAME} [xz]', t={'apk'}),
    }

  async def _engage_tamper_discard(self, args: deque[str]) -> None:
    apk = self._helper.require_target()

    _ = args.popleft()

    import time

    at = time.time()

    context = await self._helper.get_context().require_type('apk').analyze(level=2)
    with context.store().query().scoped() as q:
      if not q.patch_exists(None):
        ui.fatal('nothing to discard')
      ui.info('discarding patches to {apk}'.format(apk=apk))
      q.patch_clear()

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  async def _engage_tamper_apply(self, args: deque[str]) -> None:
    apk = self._helper.require_target()

    cmd = args.popleft()

    import os
    import time
    from tempfile import TemporaryDirectory
    from trueseeing.core.android.asm import APKAssembler
    from trueseeing.core.android.tools import move_apk

    if apk.endswith('.xapk'):
      ui.warn('patching xapk as merged apk')
      apk = apk.replace('.xapk', '.apk')

    origapk = apk.replace('.apk', '.apk.orig')

    if os.path.exists(origapk) and not cmd.endswith('!'):
      ui.fatal('backup file exists; force (!) to overwrite')

    at = time.time()

    context = await self._helper.get_context().require_type('apk').analyze(level=2)
    with context.store().query().scoped() as q:
      if not q.patch_exists(None):
        ui.fatal('nothing to apply')

      with TemporaryDirectory(dir=context.wd) as td:
        ui.info('applying patches to {apk}'.format(apk=apk))
        root = os.path.join(td, 'f')

        for path,blob in q.file_enum(None, patched=True):
          target = os.path.join(root, *path.split('/'))
          os.makedirs(os.path.dirname(target), exist_ok=True)
          with open(target, 'wb') as f:
            f.write(blob)

        outapk, outsig = await APKAssembler.assemble_from_path(td, root)

        if os.path.exists(apk):
          move_apk(apk, origapk)

        move_apk(outapk, apk)

      q.patch_clear()

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  async def _engage_tamper_disable_pinning(self, args: deque[str]) -> None:
    apk = self._helper.require_target()

    _ = args.popleft()

    import time
    import random
    from importlib.resources import files

    ui.info('disabling declarative TLS pinning {apk}'.format(apk=apk))

    at = time.time()
    context = await self._helper.get_context().require_type('apk').analyze(level=2)
    with context.store().query().scoped() as q:
      key = 'nsc{:04x}'.format(random.randint(0, 2**16))

      path = 'AndroidManifest.xml'
      blob = q.file_get(path, patched=True)
      assert blob is not None

      manif = self._parsed_manifest(blob)
      for e in manif.xpath('.//application'):
        e.attrib['{http://schemas.android.com/apk/res/android}usesCleartextTraffic'] = "true"
        e.attrib['{http://schemas.android.com/apk/res/android}networkSecurityConfig'] = f'@xml/{key}'
      q.patch_put(path, self._manifest_as_xml(manif), z=True)

      # XXX
      path = f'resources/package_1/res/xml/{key}.xml'
      q.patch_put(path, (files('trueseeing')/'libs'/'android'/'nsc.xml').read_bytes(), z=True)

      # XXX
      import lxml.etree as ET
      path = 'resources/package_1/res/values/public.xml'
      root = ET.fromstring(q.file_get(path, patched=True), parser=ET.XMLParser(recover=True))
      if root.xpath('./public[@type="xml"]'):
        maxid = max(int(e.attrib["id"], 16) for e in root.xpath('./public[@type="xml"]'))
        n = ET.SubElement(root, 'public')
        n.attrib['id'] = f'0x{maxid+1:08x}'
        n.attrib['type'] = 'xml'
        n.attrib['name'] = key
      else:
        maxid = (max(int(e.attrib["id"], 16) for e in root.xpath('./public')) & 0xffff0000)
        n = ET.SubElement(root, 'public')
        n.attrib['id'] = f'0x{maxid+0x10000:08x}'
        n.attrib['type'] = 'xml'
        n.attrib['name'] = key
      q.patch_put(path, ET.tostring(root),z=True)

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  async def _engage_tamper_enable_debug(self, args: deque[str]) -> None:
    apk = self._helper.require_target()

    _ = args.popleft()

    import time

    ui.info('enabling debug {apk}'.format(apk=apk))

    at = time.time()
    context = await self._helper.get_context().require_type('apk').analyze(level=2)
    with context.store().query().scoped() as q:
      path = 'AndroidManifest.xml'
      blob = q.file_get(path, patched=True)
      assert blob is not None
      manif = self._parsed_manifest(blob)
      for e in manif.xpath('.//application'):
        e.attrib['{http://schemas.android.com/apk/res/android}debuggable'] = "true"
      q.patch_put(path, self._manifest_as_xml(manif), z=True)

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  async def _engage_tamper_enable_backup(self, args: deque[str]) -> None:
    apk = self._helper.require_target()

    _ = args.popleft()

    import time

    ui.info('enabling full backup {apk}'.format(apk=apk))

    at = time.time()
    context = await self._helper.get_context().require_type('apk').analyze(level=1)
    with context.store().query().scoped() as q:
      path = 'AndroidManifest.xml'
      blob = q.file_get(path, patched=True)
      assert blob is not None
      manif = self._parsed_manifest(blob)
      for e in manif.xpath('.//application'):
        e.attrib['{http://schemas.android.com/apk/res/android}allowBackup'] = "true"
        if '{http://schemas.android.com/apk/res/android}fullBackupContent' in e.attrib:
          del e.attrib['{http://schemas.android.com/apk/res/android}fullBackupContent']
      q.patch_put(path, self._manifest_as_xml(manif), z=True)

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  async def _engage_tamper_patch_target_api_level(self, args: deque[str]) -> None:
    apk = self._helper.require_target()

    cmd = args.popleft()

    try:
      level = int(args.popleft())
    except (IndexError, ValueError):
      ui.fatal('need API level')

    import time

    ui.info('retargetting API level {level} {apk}'.format(level=level, apk=apk))

    at = time.time()

    context = await self._helper.get_context().require_type('apk').analyze(level=2)
    with context.store().query().scoped() as q:
      path = 'AndroidManifest.xml'
      blob = q.file_get(path, patched=True)
      assert blob is not None
      manif = self._parsed_manifest(blob)
      for e in manif.xpath('.//uses-sdk'):
        e.attrib['{http://schemas.android.com/apk/res/android}targetSdkVersion'] = str(level)
        minLevel = int(e.attrib.get('{http://schemas.android.com/apk/res/android}minSdkVersion', '1'))
        if level < minLevel:
          if not cmd.endswith('!'):
            ui.fatal('cannot target API level below requirement ({minlv}); force (!) to downgrade altogether'.format(minlv=minLevel))
          else:
            ui.warn('downgrading the requirement')
            e.attrib['{http://schemas.android.com/apk/res/android}minSdkVersion'] = str(level)
      q.patch_put(path, self._manifest_as_xml(manif), z=True)

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  # XXX: long and ugly
  async def _engage_tamper_inject_frida(self, args: deque[str], script_dir_mode: bool = False) -> None:
    configfn: Optional[str] = None
    config_override: Optional[str] = None
    dev_frida_dir: Optional[str] = None

    apk = self._helper.require_target()

    cmd = args.popleft()

    vers = None
    for optname, optvalue in self._helper.get_effective_options(self._helper.get_modifiers(args)).items():
      if optname == 'vers':
        vers = optvalue
    if not vers:
      try:
        vers = await self._determine_recent_frida_gadget()
      except InvalidResponseError:
        ui.fatal('cannot determine recent frida-gadget version (try @o:vers=X.Y.Z)')

    assert vers is not None

    if args:
      if not args[0].startswith('@'):
        if not script_dir_mode:
          configfn = args.popleft()
        else:
          dev_frida_dir = args.popleft()

    import json
    import os
    import re
    import random
    import time
    import lief
    import tempfile
    from importlib.resources import files, as_file
    from jinja2 import Environment, FileSystemLoader

    ui.info('injecting frida-gadget {apk} [{vers}]'.format(apk=apk, vers=vers))

    at = time.time()
    gadget_map = await self._prepare_frida_gadget(vers=vers)
    context: APKContext = await self._helper.get_context().require_type('apk').analyze(level=3)

    if script_dir_mode:
      from trueseeing.core.env import get_device_frida_dir

      if not dev_frida_dir:
        dev_frida_dir = get_device_frida_dir(context.get_package_name())

      assert dev_frida_dir
      ui.info(f'using frida script directory: {dev_frida_dir}')

      with as_file(files('trueseeing')/'libs'/'android') as libpath:
        env = Environment(loader=FileSystemLoader(libpath), autoescape=True, trim_blocks=True, lstrip_blocks=True)
        config_override = env.get_template('frida-scriptdir.config').render(path=dev_frida_dir)

    with context.store().query().scoped() as q:
      thunk_pat = r'p[0-9a-f]{4}\.App[0-9a-f]{8}'
      pkg = 'p{:04x}'.format(random.randint(0, 2**16 - 1))
      key = 'App{:08x}'.format(random.randint(0, 2**32 - 1))
      lib = '{:08x}'.format(random.randint(0, 2**32 - 1))

      ui.info('adding gadget{}'.format('/config' if (configfn or config_override) else ''))
      store_map_fn = 'uncompressed-files.json'
      store_map_json = q.file_get(store_map_fn, patched=True)
      assert store_map_json
      store_map = json.loads(store_map_json)
      should_store = any([l.startswith('lib/') for l in store_map['paths']])
      for arch, distfn in gadget_map.items():
        from lzma import decompress
        memberfn = f'lib/{arch}/lib{lib}.so'
        with open(distfn, 'rb') as f:
          q.patch_put(f'root/{memberfn}', decompress(f.read()), z=True)
        if should_store:
          store_map['paths'].append(memberfn)

        if configfn or config_override:
          memberfn = f'lib/{arch}/lib{lib}.config.so'
          if config_override:
            config = config_override
          else:
            assert configfn # XXX
            with open(configfn, 'r') as f:
              config = f.read()
          q.patch_put(f'root/{memberfn}', config.encode('utf-8'), z=True)
          if should_store:
            store_map['paths'].append(memberfn)

      if should_store:
        q.patch_put(store_map_fn, json.dumps(store_map).encode('utf-8'), z=True)

      ui.info('patching ELF')
      with tempfile.TemporaryDirectory(dir=context.wd) as td:
        for n, c in q.file_enum(r'^root/lib/lib.*\.so', patched=True, regex=True):
          sopath = os.path.join(td, os.path.basename(n))
          with open(sopath, 'wb') as f0:
            f0.write(c)
          so = lief.ELF.parse(sopath)
          if not so:
            ui.warn(f'file is not ELF, ignored: {n}')
          else:
            so.add_library(f'lib{lib}.so')
            so.write(sopath)
            with open(sopath, 'rb') as f1:
              q.patch_put(n, f1.read(), z=True)

      ui.info('patching dex')

      path = 'AndroidManifest.xml'
      blob = q.file_get(path, patched=True)
      assert blob is not None
      manif = self._parsed_manifest(blob)

      try:
        last_block = max([int(n) for n in set([re.match(r'smali/classes([0-9]+)', fn).group(1) for fn in q.file_find('^smali/classes[0-9]+', regex=True)])]) # type: ignore[union-attr]
        ui.info(f'detected last dex block id: {last_block}')
      except ValueError:
        ui.warn('cannot determine last dex block id, assuming single block')
        last_block = None

      ns = dict(android='http://schemas.android.com/apk/res/android')
      ns_android = ns['android']

      if not manif.xpath('./uses-permission[@android:name="android.permission.INTERNET"]', namespaces=ns):
        ui.warn('adding necessary permission: INTERNET')
        import lxml.etree as ET
        el = ET.Element('uses-permission')
        el.attrib[f'{{{ns_android}}}name'] = 'android.permission.INTERNET'
        manif.insert(0, el)

      cn = f'{pkg}.{key}'
      for e in manif.xpath('.//application'):
        base = e.attrib.get(f'{{{ns_android}}}name', 'android.app.Application')
        e.attrib[f'{{{ns_android}}}name'] = cn

        if configfn or config_override:
          ui.warn('enabling legacy behavior on native library packaging')
          e.attrib[f'{{{ns_android}}}extractNativeLibs'] = 'true'

        if not base.startswith('android.'):
          if re.fullmatch(thunk_pat, base):
            if not cmd.endswith('!'):
              ui.fatal('target appears to already injected; force (!) to do anyway')

          appcl: Dict[str, bytes] = dict()
          for n, c in q.file_enum(self._as_smali_file_pattern(base), patched=True, regex=True):
            appcl[n] = c
          if len(appcl) != 1:
            if len(appcl):
              ui.warn('found multiple candidates of application classses')
            else:
              ui.fatal('application class not found: {}'.format(base))
          for n, c in appcl.items():
            m = re.search(rb'^\.class .*? final .*?(L.+?;)', c)
            if m:
              ui.warn('the application class seems final, de-finalizing: {cl}'.format(cl=m.group(1).decode('utf-8', errors='replace')))
              q.patch_put(n, re.sub(rb'^(\.class .*?) final ', rb'\1 ', c), z=True)

      q.patch_put(path, self._manifest_as_xml(manif), z=True)

      path = 'smali/classes{last_block}/{pkg}/{key}.smali'.format(
        last_block=str(last_block) if last_block else '',
        pkg=pkg,
        key=key,
      )
      with as_file(files('trueseeing')/'libs'/'android') as libpath:
        env = Environment(loader=FileSystemLoader(libpath), autoescape=True, trim_blocks=True, lstrip_blocks=True)
        q.patch_put(
          path,
          env.get_template('frida-app.smali').render(
            base=self._as_dalvik_classname(base),
            cn=self._as_dalvik_classname(cn),
            lib=lib,
          ).encode('utf-8'),
          z=True
        )

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  async def _engage_tamper_inject_frida_scriptdir(self, args: deque[str]) -> None:
    await self._engage_tamper_inject_frida(args, script_dir_mode=True)

  def _as_dalvik_classname(self, jn: str) -> str:
    return 'L{};'.format(jn.replace('.', '/'))

  def _as_smali_file_pattern(self, jn: str) -> str:
    return r'^smali/classes[0-9]+/{}.smali$'.format(jn.replace('.', '/'))

  async def _prepare_frida_gadget(self, vers: str) -> Mapping[str, str]:
    import os
    from aiohttp import ClientSession
    from aiohttp.client_exceptions import ClientConnectionError
    context = self._helper.get_context()
    feedtmpl = 'https://github.com/frida/frida/releases/download/{vers}/frida-gadget-{vers}-android-{arch}.so.xz'
    pathtmpl = '{wd}/../t/frida-gadget-{vers}-android-{arch}.so.xz'
    archmap = dict(arm64='arm64-v8a', arm='armeabi-v7a', x86='x86', x86_64='x86_64')

    o = dict()

    async with ClientSession() as sess:
      for da, aa in archmap.items():
        feed = feedtmpl.format(vers=vers, arch=da)
        path = pathtmpl.format(wd=context.wd, vers=vers, arch=da)
        if not os.path.exists(path):
          ui.info(f'fetching {feed}')
          os.makedirs(os.path.dirname(path), exist_ok=True)
          try:
            async with sess.get(feed) as r:
              with open(path, 'wb') as f:
                if r.status != 200:
                  ui.warn('failed to fetch {feed}: {status}'.format(feed=feed, status=r.status))
                f.write(await r.read())
          except ClientConnectionError as e:
            ui.fatal(f'failed to fetch {feed}: {e}')
        o[aa] = path

    return o

  async def _determine_recent_frida_gadget(self) -> str:
    import re
    from aiohttp import ClientSession
    from aiohttp.client_exceptions import ClientConnectionError
    feed = 'https://frida.re/news/releases/'

    async with ClientSession() as sess:
      try:
        async with sess.get(feed) as r:
          m = re.search(r'<article>.*?frida ([0-9.]+).*</article>', await r.text(), re.DOTALL | re.IGNORECASE)
          if m:
            return m.group(1)
          else:
            raise InvalidResponseError()
      except ClientConnectionError:
        raise InvalidResponseError()

  async def _engage_device_copyout(self, args: deque[str]) -> None:
    success: bool = False

    cmd = args.popleft()
    if not args:
      ui.fatal('need package name')

    target = args.popleft()

    import os
    if not args:
      outfn = f'{target}.tar'
    else:
      outfn = args.popleft()

    outfn0 = outfn.replace('.tar', '') + '-int.tar'
    outfn1 = outfn.replace('.tar', '') + '-ext.tar'

    if os.path.exists(outfn) and not cmd.endswith('!'):
      ui.fatal('outfile exists; force (!) to overwrite')

    ui.info(f'copying out: {target} -> {outfn}')

    import time
    from shlex import quote
    from subprocess import CalledProcessError
    from trueseeing.core.tools import invoke_passthru
    from trueseeing.core.android.device import AndroidDevice
    from trueseeing.core.android.tools import toolchains

    at = time.time()
    dev = AndroidDevice()

    if not await dev.is_fullbackup_available():
      ui.warn('full backup feature is not available')
    else:
      ui.info('initiating a backup on device; give "1" as the password if asked')
      await dev.invoke_adb_passthru(f'backup -f {outfn}.ab {target}')
      try:
        try:
          with toolchains() as tc:
            await invoke_passthru('java -jar {abe} unpack {outfn}.ab {outfn} 1'.format(
              abe=tc['abe'],
              outfn=quote(outfn),
            ))
        except CalledProcessError:
          ui.warn('unpack failed (did you give the correct password?); trying the next method')
        else:
          ui.success('unpack success')
          if os.stat(outfn).st_size > 1024:
            ui.success(f'copied out: {outfn}')
            success = True
          else:
            ui.warn('got an empty backup; trying the next method')
            try:
              os.remove(outfn)
            except FileNotFoundError:
              pass
      finally:
        try:
          os.remove(f'{outfn}.ab')
        except FileNotFoundError:
          pass

    if not success:
      if not await dev.is_package_debuggable(target):
        ui.warn('target is not debuggable')
      else:
        ui.info('target seems debuggable; trying extraction with debug interface')

        tfn0 = self._generate_tempfilename_for_device()
        ui.info('copying internal storage')
        await dev.invoke_adb_passthru(f"shell 'run-as {target} tar -cv . > {tfn0}'")
        await dev.invoke_adb_passthru(f'pull {tfn0} {quote(outfn0)}')
        await dev.invoke_adb_passthru(f'shell rm -f {tfn0}')
        ui.success(f'copied out: {outfn0}')

        ui.info('copying external storage')
        tfn1 = self._generate_tempfilename_for_device()
        try:
          await dev.invoke_adb_passthru(f"shell 'cd /storage/emulated/0/Android/ && tar -cv data/{target} obb/{target} > {tfn1}'")
        except CalledProcessError:
          ui.warn('detected errors during extraction from external storage (may indicate partial extraction)')
        await dev.invoke_adb_passthru(f'pull {tfn1} {quote(outfn1)}')
        await dev.invoke_adb_passthru(f'shell rm -f {tfn1}')
        ui.success(f'copied out: {outfn1}')

        success = True

    if success:
      ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))
    else:
      ui.failure('copyout failed')

  async def _engage_device_copyin(self, args: deque[str]) -> None:
    success: bool = False

    _ = args.popleft()
    if not args:
      ui.fatal('need package name')

    target = args.popleft()

    import os
    if not args:
      fn = f'{target}.tar'
    else:
      fn = args.popleft()

    fn0 = fn.replace('.tar', '') + '-int.tar'
    fn1 = fn.replace('.tar', '') + '-ext.tar'

    if not any(os.path.exists(x) for x in [fn, fn0, fn1]):
      ui.fatal('bundle file not found')

    ui.info(f'copying in: {fn} -> {target}')

    import time
    from shlex import quote
    from subprocess import CalledProcessError
    from trueseeing.core.tools import invoke_passthru
    from trueseeing.core.android.device import AndroidDevice
    from trueseeing.core.android.tools import toolchains

    at = time.time()
    dev = AndroidDevice()

    if not await dev.is_fullbackup_available():
      ui.warn('full backup feature is not available')
    else:
      if not os.path.exists(fn):
        ui.warn(f'data not found, trying the next method: {fn}')
      else:
        try:
          try:
            with toolchains() as tc:
              await invoke_passthru('java -jar {abe} pack-kk {fn} {fn}.ab 1'.format(
                abe=tc['abe'],
                fn=quote(fn),
              ))
          except CalledProcessError:
            ui.warn('pack failed; trying the next method')
          else:
            ui.success('pack success')
            ui.info('initiating a restore on device; give "1" as the password if asked')
            await dev.invoke_adb_passthru(f'restore {quote(fn)}.ab')
            ui.success(f'copied in: {fn}')
            success = True
        finally:
          try:
            os.remove(f'{fn}.ab')
          except FileNotFoundError:
            pass

    if not success:
      if not await dev.is_package_debuggable(target):
        ui.warn('target is not debuggable')
      else:
        ui.info('target seems debuggable; trying injection with debug interface')

        ui.info('copying internal storage')
        if not os.path.exists(fn0):
          ui.warn(f'data not found: {fn0}')
        else:
          tfn0 = self._generate_tempfilename_for_device()
          await dev.invoke_adb_passthru(f'push {quote(fn0)} {tfn0}')
          await dev.invoke_adb_passthru(f"shell 'run-as {target} tar -xv < {tfn0}; rm -f {tfn0}'")
          ui.success(f'copied in: {fn}')
          success = True

        ui.info('copying external storage')
        if not os.path.exists(fn1):
          ui.warn(f'data not found: {fn1}')
        else:
          tfn1 = self._generate_tempfilename_for_device()
          await dev.invoke_adb_passthru(f'push {quote(fn1)} {tfn1}')
          await dev.invoke_adb_passthru(f"shell 'cd /storage/emulated/0/Android/ && tar -xv < {tfn1}; rm -f {tfn1}'")
          ui.success(f'copied in: {fn1}')
          success = True

        success = True

    if success:
      ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))
    else:
      ui.failure('copyin failed')

  async def _engage_fuzz_command(self, args: deque[str], am: bool = False) -> None:
    outfn: Optional[str] = None

    cmd = args.popleft()

    if not args:
      if am:
        ui.fatal('an "am" command line pattern required; try giving whatever you would to "adb shell am" (e.g. {} "start-activity .." ..)'.format(cmd))
      else:
        ui.fatal('command line pattern required; try giving you would to "adb shell"')

    pat = args.popleft()
    if am:
      pat = f'am {pat}'

    if args and not args[0].startswith('@'):
      import os
      outfn = args.popleft()
      if os.path.exists(outfn) and not cmd.endswith('!'):
        ui.fatal('outfile exists; force (!) to overwrite')

    wordlist: Dict[str, List[str]] = dict()
    for name, fn in self._helper.get_effective_options(self._helper.get_modifiers(args)).items():
      if name.startswith('w'):
        name = name[1:]
        try:
          with open(fn, 'r') as f:
            wordlist[name] = [x.rstrip() for x in f]
        except OSError as e:
          ui.fatal(f'cannot open wordlist: {e}')

    if not wordlist:
      ui.fatal('need a wordlist (try @o:wNAME=FN)')

    ui.info('wordlist built: {} words in {} keys ({})'.format(sum([len(v) for v in wordlist.values()]), len(wordlist), ','.join(wordlist.keys())))

    def _expand(pat: str, wordlist: Mapping[str, List[str]]) -> Iterator[Tuple[int, int, str]]:
      tries = min(len(v) for v in wordlist.values())
      for nr in range(tries):
        d = {k:v[nr] for k,v in wordlist.items()}
        try:
          yield nr, tries, pat.format(*[], **d)
        except KeyError as e:
          ui.fatal(f'unknown wordlist specified: {e}')

    ui.info('starting fuzzing, opening log system-wide{}'.format(' [{}]'.format(outfn) if outfn else ''))

    from trueseeing.core.android.device import AndroidDevice

    dev = AndroidDevice()

    async def _log(outfn: Optional[str]) -> None:
      import sys
      nr = 0

      if not outfn:
        f = sys.stdout.buffer
      else:
        f = open(outfn, 'wb')

      try:
        async for l in dev.invoke_adb_streaming('logcat -T1'):
          f.write(l)
          nr += 1
          if outfn and nr % 256 == 0:
            ui.info(' ... captured: {}')
      finally:
        if outfn:
          f.close()

    async def _fuzz(pat: str, wordlist: Mapping[str, List[str]]) -> None:
      from asyncio import sleep
      from subprocess import CalledProcessError
      for nr, tries, t in _expand(pat, wordlist):
        await sleep(.05)
        prog = dict(nr=nr+1, max=tries, cmd=t)
        try:
          await dev.invoke_adb(f'shell {t}')
          ui.info('[{nr}/{max}] {cmd}'.format(**prog))
        except CalledProcessError as e:
          ui.failure('[{nr}/{max}] {cmd}: failed: {code}'.format(code=e.returncode, **prog))

    from asyncio import create_task, wait, FIRST_COMPLETED, ALL_COMPLETED
    task_log = create_task(_log(outfn))
    task_fuzz = create_task(_fuzz(pat, wordlist))

    done, pending = await wait([task_log, task_fuzz], return_when=FIRST_COMPLETED)
    for t in pending:
      t.cancel()
    done, _ = await wait([task_log, task_fuzz], return_when=ALL_COMPLETED)
    for t in done:
      exc = t.exception()
      if exc:
        ui.error('unhandled exception', exc=exc)

  async def _engage_fuzz_intent(self, args: deque[str]) -> None:
    await self._engage_fuzz_command(args, am=True)

  async def _engage_deploy_package(self, args: deque[str]) -> None:
    cmd = args.popleft()

    context: APKContext = self._helper.get_context().require_type('apk')
    apk = context.target

    from time import time
    from shlex import quote
    from pubsub import pub
    from trueseeing.core.ui import AndroidInstallProgressReporter
    from trueseeing.core.android.device import AndroidDevice
    from subprocess import CalledProcessError

    dev = AndroidDevice()
    at = time()
    pkg = context.get_package_name()

    ui.info(f'deploying package: {pkg}')

    if cmd.endswith('!'):
      try:
        async for l in dev.invoke_adb_streaming(f'uninstall {pkg}', redir_stderr=True):
          pub.sendMessage('progress.android.adb.update')
          if b'success' in l.lower():
            ui.warn('removing existing package')
      except CalledProcessError as e:
        ui.fatal('uninstall failed: {}'.format(e.stdout.decode().rstrip()))

    with AndroidInstallProgressReporter().scoped():
      pub.sendMessage('progress.android.adb.begin', what='installing ... ')
      try:
        async for l in dev.invoke_adb_streaming(f'install --no-streaming {quote(apk)}', redir_stderr=True):
          pub.sendMessage('progress.android.adb.update')
          if b'failure' in l.lower():
            ui.stderr('')
            if not cmd.endswith('!'):
              ui.fatal('install failed; force (!) to replace ({})'.format(l.decode('UTF-8')))
            else:
              ui.fatal('install failed ({})'.format(l.decode('UTF-8')))

        pub.sendMessage('progress.android.adb.done')
      except CalledProcessError as e:
        ui.fatal('install failed: {}'.format(e.stdout.decode().rstrip()))

      ui.success('done ({t:.02f} sec){trailer}'.format(t=time() - at, trailer=' '*8))

  async def _engage_undeploy_package(self, args: deque[str]) -> None:
    _ = args.popleft()

    context: APKContext = self._helper.get_context().require_type('apk')

    from time import time
    from pubsub import pub
    from trueseeing.core.android.device import AndroidDevice
    from subprocess import CalledProcessError

    dev = AndroidDevice()
    at = time()
    pkg = context.get_package_name()

    ui.info(f'removing package: {pkg}')

    try:
      async for l in dev.invoke_adb_streaming(f'uninstall {pkg}', redir_stderr=True):
        pub.sendMessage('progress.android.adb.update')
        if b'failure' in l.lower():
          import re
          packages = await dev.invoke_adb('shell pm list packages', redir_stderr=True)
          if not re.match(f'{pkg}$', packages, re.MULTILINE):
            ui.fatal('package not found')
          else:
            ui.fatal('uninstall failed ({})'.format(l.decode()))
    except CalledProcessError as e:
      ui.fatal('uninstall failed: {}'.format(e.stdout.decode().rstrip()))

    ui.success('done ({t:.02f} sec)'.format(t=time() - at))

  async def _engage_grab_package(self, args: deque[str]) -> None:
    cmd = args.popleft()

    import os
    if not args:
      ui.fatal('need the package name')

    pkg = args.popleft()

    if args:
      outfn = args.popleft()
    else:
      outfn = f'{pkg}.apk'

    if os.path.exists(outfn):
      if not cmd.endswith('!'):
        ui.fatal('output file exists; force (!) to overwrite')
      else:
        os.remove(outfn)

    import re
    from time import time
    from tempfile import TemporaryDirectory
    from pubsub import pub
    from trueseeing.core.android.device import AndroidDevice

    dev = AndroidDevice()
    at = time()
    outfn = os.path.realpath(outfn)

    ui.info(f'grabbing package: {pkg} -> {outfn}')

    basepath: Optional[bytes] = None
    splits: List[bytes] = []

    async for l in dev.invoke_adb_streaming(f'shell pm dump {pkg}', redir_stderr=True):
      pub.sendMessage('progress.android.adb.update')
      if f'unable to find package: {pkg}'.encode() in l.lower():
        ui.fatal(f'package not found: {pkg}')

      m = re.search(rb'codePath=(/.+)', l)
      if m:
        basepath = m.group(1)
      m = re.search(rb'splits=\[(.+)\]', l)
      if m:
        splits = re.split(rb', *', m.group(1))

    assert basepath
    assert splits

    with TemporaryDirectory() as td:
      from os import chdir, getcwd
      from shlex import quote
      from zipfile import ZipFile, ZIP_STORED
      cd = getcwd()
      try:
        chdir(td)
        slicemap = dict()
        if len(splits) == 1:
          if outfn.endswith('.xapk'):
            ui.warn('target has only one slice; using apk format')
            outfn = outfn.replace('.xapk', '.apk')
          ui.info('getting {nr} slice'.format(nr=len(splits)))
          await dev.invoke_adb('pull {path}/base.apk {outfn}'.format(path=quote(basepath.decode()), outfn=quote(outfn)))
        else:
          if outfn.endswith('.apk'):
            ui.warn('target has multiple slices; using xapk format')
            outfn = outfn.replace('.apk', '.xapk')
          ui.info('getting {nr} slices'.format(nr=len(splits)))
          for s in splits:
            slice = s.decode()
            if slice == 'base':
              fn = f'{pkg}.apk'
            else:
              fn = f'{slice}.apk'
            await dev.invoke_adb('pull {path}/{typ}{slice}.apk {fn}'.format(
              path=quote(basepath.decode()),
              typ='' if slice == 'base' else 'split_',
              slice=slice,
              fn=fn,
            ))
            slicemap[slice] = fn
          XAPKManifestGenerator(slicemap).generate()
          with ZipFile(outfn, 'w', ZIP_STORED) as zf:
            from glob import glob
            for n in glob('*'):
              with open(n, 'rb') as g:
                zf.writestr(n, g.read())
      finally:
        chdir(cd)
    ui.success('done ({t:.02f} sec)'.format(t=time() - at))

  def _generate_tempfilename_for_device(self, dir: Optional[str] = None) -> str:
    import random
    return (f'{dir}/' if dir is not None else '/data/local/tmp/') + ''.join(random.choices('abcdefghijklmnopqrstuvwxyz0123456789', k=16))

  def _parsed_manifest(self, blob: bytes) -> Any:
    import lxml.etree as ET
    return ET.fromstring(blob, parser=ET.XMLParser(recover=True))

  def _manifest_as_xml(self, manifest: Any) -> bytes:
    import lxml.etree as ET
    assert manifest is not None
    return ET.tostring(manifest) # type: ignore[no-any-return]

  async def _engage_frida_start_server(self, args: deque[str]) -> None:
    cmd = args.popleft()
    context: APKContext = self._helper.get_context().require_type('apk')

    pkg = context.get_package_name()
    scripts = list(args)

    force = cmd.endswith('!')

    ui.info("starting frida-server")

    from time import time
    from shlex import quote
    from trueseeing.core.android.device import AndroidDevice
    from trueseeing.core.tools import invoke_passthru
    from pathlib import Path

    at = time()
    dev = AndroidDevice()

    if force:
      ui.warn("killing frida-server if any")
      await dev.invoke_adb("shell 'killall frida-server || exit 0'")
    await dev.invoke_adb("shell 'cd /data/local/tmp && ./frida-server &'")

    ui.info(f"starting frida on {pkg}")
    scripts_str = []
    for s in scripts:
      p = Path(s)
      if p.is_file():
        scripts_str.append(f"-l {quote(str(p))}")
      elif p.is_dir():
        scripts_str.extend([f"-l {quote(str(m))}" for m in p.rglob('*.js')])
      else:
        ui.warn(f"ignoring unknown path: {p}")
    await invoke_passthru(f"frida -U {pkg} {' '.join(scripts_str)}")

    ui.success("done ({t:.2f} sec.)".format(t=time() - at))

  async def _engage_frida_kill_server(self, args: deque[str]) -> None:
    from time import time
    from trueseeing.core.android.device import AndroidDevice

    ui.info("killing frida-server")

    at = time()
    dev = AndroidDevice()

    await dev.invoke_adb('shell killall frida-server')

    ui.success("done ({t:.2f} sec.)".format(t=time() - at))

class InvalidResponseError(Exception):
  pass

class XAPKManifestGenerator:
  def __init__(self, slicemap: Dict[str, str]) -> None:
    self._slicemap = slicemap

  def generate(self) -> None:
    from os import stat
    from pyaxmlparser import APK
    manif: XAPKManifest = dict(
      xapk_version='2',
      total_size=0,
      locales_name=dict(),
      split_apks=[],
    )

    for slice, fn in self._slicemap.items():
      conf = slice.split('.')[-1]
      manif['total_size'] += stat(fn).st_size
      manif['split_apks'].append(dict(id=slice, file=fn))
      if slice == 'base':
        apk = APK(fn)
        manif.update(dict(
          name=apk.get_app_name(),
          icon='icon.png',
          package_name=apk.get_package(),
          version_code=apk.version_code,
          version_name=apk.version_name,
          min_sdk_version=apk.get_min_sdk_version(),
          target_sdk_version=apk.get_target_sdk_version(),
          permissions=apk.get_permissions(),
        ))
        with open('icon.png', 'wb') as f:
          f.write(apk.icon_data)
      elif len(conf) == 2:
        apk = APK(fn)
        country_code = conf
        manif['locales_name'].update({
          country_code:apk.get_app_name(),
        })
      if manif['locales_name']:
        ln: Dict[str, str] = manif['locales_name']
        for k in ln.keys():
          if not ln[k]:
            ln[k] = manif['name']
    with open('manifest.json', 'w') as g:
      from json import dumps
      g.write(dumps(manif, separators=(',', ':')))



from __future__ import annotations
from typing import TYPE_CHECKING

from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Any
  from trueseeing.api import CommandHelper, Command, CommandMap

class InfoCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return InfoCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      'i':dict(e=self._info, n='i[i]', d='print info (ii: more info)'),
      'ii':dict(e=self._info2),
    }

  def _read_field(self, x: Any) -> Any:
    boolmap = {True:'yes',False:'no','true':'yes','false':'no',1:'yes',0:'no', None:'?'}
    return boolmap.get(x, x)

  async def _info(self, args: deque[str], extended: bool = False) -> None:
    target = self._helper.require_target()

    _ = args.popleft()

    analysisguidemap = {
      0: 'try a;i for more info',
      1: 'try a;i for more info',
      2: 'try aa;i for more info',
      3: 'try aaa;i for more info',
    }

    context = self._helper.get_context()
    analyzed = context.get_analysis_level()

    ui.info(f'info on {target}')

    async for m in context._get_info(extended):
      for k, v in m.items():
        if v is None:
          continue
        if not k.startswith('_'):
          ui.info('{:12s} {}'.format(k, self._read_field(v)))
        elif k == '_patch':
          ui.info('{:12s} {}'.format('has patch?', self._read_field(v)))
        elif k == '_analysis':
          ui.info('{:12s} {}{}'.format(
            'analyzed?',
            self._helper.decode_analysis_level(analyzed),
            ' ({})'.format(analysisguidemap[analyzed]) if analyzed < 4 else '',
          ))

  async def _info2(self, args: deque[str]) -> None:
    return await self._info(args, extended=True)



from __future__ import annotations
from typing import TYPE_CHECKING

import os
import os.path

from pubsub import pub

from trueseeing.core.env import get_home_dir
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Tuple, Optional
  from trueseeing.core.context import Context
  from trueseeing.core.db import FileEntry

class APKDisassembler:
  _context: Context

  def __init__(self, context: Context, target: Optional[str] = None):
    self._context = context
    if target:
      self._target = target
    else:
      self._target = context.target

  async def disassemble(self, level: int = 4) -> None:
    await self._do(level)

  async def _do(self, level: int) -> None:
    import glob
    import shutil
    from shlex import quote
    from trueseeing.core.tools import invoke_streaming
    from trueseeing.core.android.tools import toolchains

    cwd = os.getcwd()

    pub.sendMessage('progress.core.asm.lift.begin')

    try:
      os.chdir(self._context.wd)

      with self._context.store().query().scoped() as query:
        with toolchains() as tc:
          async for l in invoke_streaming(r'java -jar {apkeditor} d -i {apk} {suppressor} -o files'.format(
              apkeditor=tc['apkeditor'],
              apk=quote(self._target),
              suppressor='-dex' if level < 3 else '',
          ), redir_stderr=True):
            pub.sendMessage('progress.core.asm.lift.update')

          os.chdir('files')

        def read_as_row(fn: str) -> FileEntry:
          pub.sendMessage('progress.core.asm.lift.update')
          with open(fn, 'rb') as f:
            return dict(path=fn, blob=f.read(), z=True)

        def should_cache(fn: str) -> bool:
          if not os.path.isfile(fn):
            return False
          if level < 2:
            return fn == 'AndroidManifest.xml'
          else:
            return True

        query.file_put_batch(read_as_row(fn) for fn in glob.glob('**', recursive=True) if should_cache(fn))
        pub.sendMessage('progress.core.asm.lift.update')
    finally:
      os.chdir(cwd)
      pub.sendMessage('progress.core.asm.lift.update')
      shutil.rmtree(os.path.join(self._context.wd, 'files'), ignore_errors=True)
      pub.sendMessage('progress.core.asm.lift.done')

  @classmethod
  async def disassemble_to_path(cls, target: str, path: str, nodex: bool = False, merge: bool = False) -> None:
    import os
    from shlex import quote
    from tempfile import TemporaryDirectory
    from trueseeing.core.tools import invoke_streaming
    from trueseeing.core.android.tools import toolchains

    pub.sendMessage('progress.core.asm.disasm.begin')

    with TemporaryDirectory() as td:
      with toolchains() as tc:
        apk = target
        if target.endswith('.xapk'):
          if not merge:
            raise ValueError('cannot disassemble xapk without merging')
          with TemporaryDirectory() as td2:
            from zipfile import ZipFile
            with ZipFile(target) as zf:
              for n in zf.namelist():
                zf.extract(n, path=td2)
            tmpapk = os.path.join(td, 'merged.apk')
            async for l in invoke_streaming(
              '(java -jar {apkeditor} m -i {path} -o {tmpapk})'.format(
                apkeditor=tc['apkeditor'],
                path=td2,
                tmpapk=tmpapk,
              )
            ):
              print(l)
          apk = tmpapk
        else:
          apk = target

      async for l in invoke_streaming(
        '(java -jar {apkeditor} d -o {path} -i {apk} {s})'.format(
          apk=quote(apk),
          s='-dex' if nodex else '',
          apkeditor=tc['apkeditor'],
          path=quote(path),
        ), redir_stderr=True
      ):
        pub.sendMessage('progress.core.asm.disasm.update')

    pub.sendMessage('progress.core.asm.disasm.done')

class APKAssembler:
  @classmethod
  async def assemble_from_path(cls, wd: str, path: str) -> Tuple[str, str]:
    import os
    from shlex import quote
    from trueseeing.core.tools import invoke_streaming
    from trueseeing.core.android.tools import toolchains

    pub.sendMessage('progress.core.asm.asm.begin')

    with toolchains() as tc:
      async for l in invoke_streaming(
        '(java -jar {apkeditor} b -i {path} -o {wd}/output.apk && java -jar {apksigner} sign --ks {keystore} --ks-pass pass:android {wd}/output.apk)'.format(
          wd=quote(wd), path=quote(path),
          apkeditor=tc['apkeditor'],
          apksigner=tc['apksigner'],
          keystore=await SigningKey().key(),
        ), redir_stderr=True
      ):
        pub.sendMessage('progress.core.asm.asm.update')

    pub.sendMessage('progress.core.asm.asm.done')

    return os.path.join(wd, 'output.apk'), os.path.join(wd, 'output.apk.idsig')

  @classmethod
  async def merge_slices(cls, xapk: str, wd: str) -> Tuple[str, str]:
    import os
    from shlex import quote
    from trueseeing.core.tools import invoke_streaming
    from trueseeing.core.android.tools import toolchains

    pub.sendMessage('progress.core.asm.asm.begin')

    with toolchains() as tc:
      async for l in invoke_streaming(
        'java -jar {apkeditor} m -i {xapk} -o {wd}/output.apk'.format(
          wd=wd, xapk=quote(xapk),
          apkeditor=tc['apkeditor'],
        ), redir_stderr=True
      ):
        pub.sendMessage('progress.core.asm.asm.update')

    pub.sendMessage('progress.core.asm.asm.done')

    return os.path.join(wd, 'output.apk'), os.path.join(wd, 'output.apk.idsig')

class SigningKey:
  _path: str

  def __init__(self) -> None:
    self._path = os.path.join(get_home_dir(), 'sign.keystore')

  async def key(self) -> str:
    os.makedirs(os.path.dirname(self._path), exist_ok=True)
    if not os.path.exists(self._path):
      await self._generate()
    return self._path

  async def _generate(self) -> None:
    from shlex import quote
    from trueseeing.core.tools import invoke_passthru
    ui.info("generating key for repackaging")
    await invoke_passthru(f'keytool -genkey -v -keystore {quote(self._path)} -alias androiddebugkey -dname "CN=Android Debug, O=Android, C=US" -storepass android -keypass android -keyalg RSA -keysize 2048 -validity 10000')



from __future__ import annotations
from typing import TYPE_CHECKING, NamedTuple

from trueseeing.api import Signature

if TYPE_CHECKING:
  from typing import Optional, TypedDict, Dict, List
  from trueseeing.api import SignatureHelper, ConfigMap
  from trueseeing.core.android.context import APKContext
  from trueseeing.core.android.analyze.op import OpAnalyzer

  class XAPKSliceMember(TypedDict):
    id: str
    file: str

  class XAPKManifest(TypedDict, total=False):
    xapk_version: str
    total_size: int
    locales_name: Dict[str, str]
    split_apks: List[XAPKSliceMember]
    name: str
    icon: str
    package_name: str
    version_code: str
    version_name: str
    min_sdk_version: str
    target_sdk_version: str
    permissions: List[str]

  class Call(TypedDict):
    path: str
    sect: str
    offs: int
    priv: bool
    cpp: bool
    target: str

class Op(NamedTuple):
  addr: int
  l: str

class Token(NamedTuple):
  t: str
  v: str

class InvocationPattern(NamedTuple):
  insn: str
  value: str
  i: Optional[int] = None

class SignatureMixin(Signature):
  __an: Optional[OpAnalyzer] = None

  def __init__(self, helper: SignatureHelper) -> None:
    self._helper = helper

  def get_configs(self) -> ConfigMap:
    return dict()

  @property
  def _an(self) -> OpAnalyzer:
    if self.__an is None:
      from trueseeing.core.android.analyze.op import OpAnalyzer
      self.__an = OpAnalyzer()
    return self.__an

  def _get_context(self) -> APKContext:
    return self._helper.get_context().require_type('apk')
